# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating Clusters via the DB API"""
from oslo_utils import uuidutils

from magnum.common import context
from magnum.common import exception
from magnum.objects.fields import ClusterStatus as cluster_status
from magnum.tests.unit.db import base
from magnum.tests.unit.db import utils


class DbClusterTestCase(base.DbTestCase):

    def test_create_cluster(self):
        utils.create_test_cluster()

    def test_create_cluster_nullable_cluster_template_id(self):
        utils.create_test_cluster(cluster_template_id=None)

    def test_create_cluster_already_exists(self):
        utils.create_test_cluster()
        self.assertRaises(exception.ClusterAlreadyExists,
                          utils.create_test_cluster)

    def test_get_cluster_by_id(self):
        cluster = utils.create_test_cluster()
        res = self.dbapi.get_cluster_by_id(self.context, cluster.id)
        self.assertEqual(cluster.id, res.id)
        self.assertEqual(cluster.uuid, res.uuid)

    def test_get_cluster_by_name(self):
        cluster = utils.create_test_cluster()
        res = self.dbapi.get_cluster_by_name(self.context, cluster.name)
        self.assertEqual(cluster.name, res.name)
        self.assertEqual(cluster.uuid, res.uuid)

    def test_get_cluster_by_uuid(self):
        cluster = utils.create_test_cluster()
        res = self.dbapi.get_cluster_by_uuid(self.context, cluster.uuid)
        self.assertEqual(cluster.id, res.id)
        self.assertEqual(cluster.uuid, res.uuid)

    def test_get_cluster_that_does_not_exist(self):
        self.assertRaises(exception.ClusterNotFound,
                          self.dbapi.get_cluster_by_id,
                          self.context, 999)
        self.assertRaises(exception.ClusterNotFound,
                          self.dbapi.get_cluster_by_uuid,
                          self.context,
                          '12345678-9999-0000-aaaa-123456789012')
        self.assertRaises(exception.ClusterNotFound,
                          self.dbapi.get_cluster_by_name,
                          self.context, 'not_found')

    def test_get_cluster_by_name_multiple_cluster(self):
        utils.create_test_cluster(
            id=1, name='clusterone',
            uuid=uuidutils.generate_uuid())
        utils.create_test_cluster(
            id=2, name='clusterone',
            uuid=uuidutils.generate_uuid())
        self.assertRaises(exception.Conflict,
                          self.dbapi.get_cluster_by_name,
                          self.context, 'clusterone')

    def test_get_all_cluster_stats(self):
        uuid1 = uuidutils.generate_uuid()
        utils.create_test_cluster(
            id=1, name='clusterone',
            uuid=uuid1)
        utils.create_nodegroups_for_cluster(cluster_id=uuid1)
        uuid2 = uuidutils.generate_uuid()
        utils.create_test_cluster(
            id=2, name='clustertwo',
            uuid=uuid2)
        utils.create_nodegroups_for_cluster(cluster_id=uuid2)
        ret = self.dbapi.get_cluster_stats(self.context)
        self.assertEqual(ret, (2, 12))

    def test_get_one_tenant_cluster_stats(self):
        uuid1 = uuidutils.generate_uuid()
        utils.create_test_cluster(
            id=1, name='clusterone', project_id='proj1',
            uuid=uuid1)
        utils.create_nodegroups_for_cluster(
            cluster_id=uuid1, project_id='proj1')
        uuid2 = uuidutils.generate_uuid()
        utils.create_test_cluster(
            id=2, name='clustertwo', project_id='proj2',
            uuid=uuid2)
        utils.create_nodegroups_for_cluster(
            cluster_id=uuid2, project_id='proj2')
        ret = self.dbapi.get_cluster_stats(self.context, 'proj2')
        self.assertEqual(ret, (1, 6))

    def test_get_cluster_list(self):
        uuids = []
        for i in range(1, 6):
            cluster = utils.create_test_cluster(uuid=uuidutils.generate_uuid())
            uuids.append(str(cluster['uuid']))
        res = self.dbapi.get_cluster_list(self.context)
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), sorted(res_uuids))

    def test_get_cluster_list_sorted(self):
        uuids = []
        for _ in range(5):
            cluster = utils.create_test_cluster(uuid=uuidutils.generate_uuid())
            uuids.append(str(cluster.uuid))
        res = self.dbapi.get_cluster_list(self.context, sort_key='uuid')
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), res_uuids)

        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.get_cluster_list,
                          self.context,
                          sort_key='foo')

    def test_get_cluster_list_with_filters(self):
        ct1 = utils.get_test_cluster_template(id=1,
                                              uuid=uuidutils.generate_uuid())
        ct2 = utils.get_test_cluster_template(id=2,
                                              uuid=uuidutils.generate_uuid())
        self.dbapi.create_cluster_template(ct1)
        self.dbapi.create_cluster_template(ct2)

        uuid1 = uuidutils.generate_uuid()
        cluster1 = utils.create_test_cluster(
            name='cluster-one',
            uuid=uuid1,
            cluster_template_id=ct1['uuid'],
            status=cluster_status.CREATE_IN_PROGRESS)
        utils.create_nodegroups_for_cluster(cluster_id=uuid1)
        uuid2 = uuidutils.generate_uuid()
        cluster2 = utils.create_test_cluster(
            name='cluster-two',
            uuid=uuid2,
            cluster_template_id=ct2['uuid'],
            status=cluster_status.UPDATE_IN_PROGRESS)
        utils.create_nodegroups_for_cluster(
            cluster_id=uuid2, node_count=1, master_count=1)
        cluster3 = utils.create_test_cluster(
            name='cluster-three',
            status=cluster_status.DELETE_IN_PROGRESS)
        utils.create_nodegroups_for_cluster(
            node_count=2, master_count=5)

        res = self.dbapi.get_cluster_list(
            self.context, filters={'cluster_template_id': ct1['uuid']})
        self.assertEqual([cluster1.id], [r.id for r in res])

        res = self.dbapi.get_cluster_list(
            self.context, filters={'cluster_template_id': ct2['uuid']})
        self.assertEqual([cluster2.id], [r.id for r in res])

        res = self.dbapi.get_cluster_list(self.context,
                                          filters={'name': 'cluster-one'})
        self.assertEqual([cluster1.id], [r.id for r in res])

        res = self.dbapi.get_cluster_list(self.context,
                                          filters={'name': 'bad-cluster'})
        self.assertEqual([], [r.id for r in res])

        res = self.dbapi.get_cluster_list(self.context,
                                          filters={'node_count': 3})
        self.assertEqual([cluster1.id], [r.id for r in res])

        res = self.dbapi.get_cluster_list(self.context,
                                          filters={'node_count': 1})
        self.assertEqual([cluster2.id], [r.id for r in res])

        res = self.dbapi.get_cluster_list(self.context,
                                          filters={'master_count': 3})
        self.assertEqual([cluster1.id], [r.id for r in res])

        res = self.dbapi.get_cluster_list(self.context,
                                          filters={'master_count': 1})
        self.assertEqual([cluster2.id], [r.id for r in res])

        # Check that both filters have to be valid
        filters = {'master_count': 1, 'node_count': 1}
        res = self.dbapi.get_cluster_list(self.context, filters=filters)
        self.assertEqual([cluster2.id], [r.id for r in res])

        filters = {'master_count': 1, 'node_count': 2}
        res = self.dbapi.get_cluster_list(self.context, filters=filters)
        self.assertEqual(0, len(res))

        filters = {'status': [cluster_status.CREATE_IN_PROGRESS,
                              cluster_status.DELETE_IN_PROGRESS]}
        res = self.dbapi.get_cluster_list(self.context,
                                          filters=filters)
        self.assertEqual([cluster1.id, cluster3.id], [r.id for r in res])

    def test_get_cluster_list_by_admin_all_tenants(self):
        uuids = []
        for i in range(1, 6):
            cluster = utils.create_test_cluster(
                uuid=uuidutils.generate_uuid(),
                project_id=uuidutils.generate_uuid(),
                user_id=uuidutils.generate_uuid())
            uuids.append(str(cluster['uuid']))
        ctx = context.make_admin_context(all_tenants=True)
        res = self.dbapi.get_cluster_list(ctx)
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), sorted(res_uuids))

    def test_get_cluster_list_cluster_template_not_exist(self):
        utils.create_test_cluster()
        self.assertEqual(1, len(self.dbapi.get_cluster_list(self.context)))
        res = self.dbapi.get_cluster_list(self.context, filters={
            'cluster_template_id': uuidutils.generate_uuid()})
        self.assertEqual(0, len(res))

    def test_destroy_cluster(self):
        cluster = utils.create_test_cluster()
        self.assertIsNotNone(self.dbapi.get_cluster_by_id(self.context,
                                                          cluster.id))
        self.dbapi.destroy_cluster(cluster.id)
        self.assertRaises(exception.ClusterNotFound,
                          self.dbapi.get_cluster_by_id,
                          self.context, cluster.id)

    def test_destroy_cluster_by_uuid(self):
        cluster = utils.create_test_cluster()
        self.assertIsNotNone(self.dbapi.get_cluster_by_uuid(self.context,
                                                            cluster.uuid))
        self.dbapi.destroy_cluster(cluster.uuid)
        self.assertRaises(exception.ClusterNotFound,
                          self.dbapi.get_cluster_by_uuid, self.context,
                          cluster.uuid)

    def test_destroy_cluster_by_id_that_does_not_exist(self):
        self.assertRaises(exception.ClusterNotFound,
                          self.dbapi.destroy_cluster,
                          '12345678-9999-0000-aaaa-123456789012')

    def test_destroy_cluster_by_uuid_that_does_not_exist(self):
        self.assertRaises(exception.ClusterNotFound,
                          self.dbapi.destroy_cluster, '999')

    def test_update_cluster(self):
        cluster = utils.create_test_cluster()
        old_status = cluster.status
        new_status = 'UPDATE_IN_PROGRESS'
        self.assertNotEqual(old_status, new_status)
        res = self.dbapi.update_cluster(cluster.id, {'status': new_status})
        self.assertEqual(new_status, res.status)

    def test_update_cluster_not_found(self):
        cluster_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ClusterNotFound, self.dbapi.update_cluster,
                          cluster_uuid, {'node_count': 5})

    def test_update_cluster_uuid(self):
        cluster = utils.create_test_cluster()
        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.update_cluster, cluster.id,
                          {'uuid': ''})



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""create trustee for each bay

Revision ID: 5d4caa6e0a42
Revises: bb42b7cad130
Create Date: 2016-02-17 14:16:12.927874

"""

# revision identifiers, used by Alembic.
revision = '5d4caa6e0a42'
down_revision = 'bb42b7cad130'

from alembic import op  # noqa: E402

from sqlalchemy.types import String  # noqa: E402

import sqlalchemy as sa  # noqa: E402


def upgrade():
    op.alter_column('bay', 'registry_trust_id',
                    new_column_name='trust_id',
                    existing_type=sa.String(255))
    op.add_column('bay', sa.Column('trustee_username',
                  String(255), nullable=True))
    op.add_column('bay', sa.Column('trustee_user_id',
                  sa.String(length=255), nullable=True))
    op.add_column('bay', sa.Column('trustee_password',
                  String(255), nullable=True))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from oslo_log import log as logging
from oslo_utils import strutils

from magnum.common import cinder
from magnum.common import exception
from magnum.common.x509 import operations as x509
from magnum.conductor.handlers.common import cert_manager
from magnum.drivers.heat import k8s_template_def
from magnum.drivers.heat import template_def
from magnum.i18n import _
from oslo_config import cfg

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class K8sFedoraTemplateDefinition(k8s_template_def.K8sTemplateDefinition):
    """Kubernetes template for a Fedora."""

    def __init__(self):
        super(K8sFedoraTemplateDefinition, self).__init__()
        self.add_parameter('docker_storage_driver',
                           cluster_template_attr='docker_storage_driver')

    def get_params(self, context, cluster_template, cluster, **kwargs):
        extra_params = kwargs.pop('extra_params', {})

        extra_params['username'] = context.user_name
        osc = self.get_osc(context)
        extra_params['region_name'] = osc.cinder_region_name()

        self._set_volumes(context, cluster, extra_params)

        extra_params['nodes_affinity_policy'] = \
            CONF.cluster.nodes_affinity_policy

        if cluster_template.network_driver == 'flannel':
            extra_params["pods_network_cidr"] = \
                cluster.labels.get('flannel_network_cidr', '10.100.0.0/16')
        if cluster_template.network_driver == 'calico':
            extra_params["pods_network_cidr"] = \
                cluster.labels.get('calico_ipv4pool', '10.100.0.0/16')

        # check cloud provider and cinder options. If cinder is selected,
        # the cloud provider needs to be enabled.
        cloud_provider_enabled = cluster.labels.get(
            'cloud_provider_enabled',
            'true' if CONF.trust.cluster_user_trust else 'false')
        if (not CONF.trust.cluster_user_trust
                and cloud_provider_enabled.lower() == 'true'):
            raise exception.InvalidParameterValue(_(
                '"cluster_user_trust" must be set to True in magnum.conf when '
                '"cloud_provider_enabled" label is set to true.'))
        if (cluster_template.volume_driver == 'cinder'
                and cloud_provider_enabled.lower() == 'false'):
            raise exception.InvalidParameterValue(_(
                '"cinder" volume driver needs "cloud_provider_enabled" label '
                'to be true or unset.'))
        extra_params['cloud_provider_enabled'] = cloud_provider_enabled

        label_list = ['coredns_tag',
                      'hyperkube_prefix',
                      'kube_tag', 'container_infra_prefix',
                      'availability_zone',
                      'cgroup_driver',
                      'container_runtime',
                      'containerd_version',
                      'containerd_tarball_url',
                      'containerd_tarball_sha256',
                      'calico_tag',
                      'calico_ipv4pool',
                      'calico_ipv4pool_ipip',
                      'cinder_csi_enabled', 'cinder_csi_plugin_tag',
                      'csi_attacher_tag', 'csi_provisioner_tag',
                      'csi_snapshotter_tag', 'csi_resizer_tag',
                      'csi_node_driver_registrar_tag',
                      'csi_liveness_probe_tag',
                      'etcd_tag', 'flannel_tag', 'flannel_cni_tag',
                      'cloud_provider_tag',
                      'prometheus_tag', 'grafana_tag',
                      'heat_container_agent_tag',
                      'keystone_auth_enabled', 'k8s_keystone_auth_tag',
                      'heapster_enabled',
                      'metrics_server_enabled',
                      'metrics_server_chart_tag',
                      'monitoring_enabled',
                      'monitoring_retention_days',
                      'monitoring_retention_size',
                      'monitoring_interval_seconds',
                      'monitoring_storage_class_name',
                      'monitoring_ingress_enabled',
                      'cluster_basic_auth_secret',
                      'cluster_root_domain_name',
                      'prometheus_operator_chart_tag',
                      'prometheus_adapter_enabled',
                      'prometheus_adapter_chart_tag',
                      'prometheus_adapter_configmap',
                      'selinux_mode',
                      'helm_client_url', 'helm_client_sha256',
                      'helm_client_tag',
                      'traefik_ingress_controller_tag',
                      'node_problem_detector_tag',
                      'nginx_ingress_controller_tag',
                      'nginx_ingress_controller_chart_tag',
                      'auto_healing_enabled', 'auto_scaling_enabled',
                      'auto_healing_controller', 'magnum_auto_healer_tag',
                      'draino_tag', 'autoscaler_tag',
                      'min_node_count', 'max_node_count', 'npd_enabled',
                      'ostree_remote', 'ostree_commit',
                      'use_podman', 'kube_image_digest',
                      'metrics_scraper_tag']

        labels = self._get_relevant_labels(cluster, kwargs)

        for label in label_list:
            label_value = labels.get(label)
            if label_value:
                extra_params[label] = label_value

        csr_keys = x509.generate_csr_and_key(u"Kubernetes Service Account")

        extra_params['kube_service_account_key'] = \
            csr_keys["public_key"].replace("\n", "\\n")
        extra_params['kube_service_account_private_key'] = \
            csr_keys["private_key"].replace("\n", "\\n")

        extra_params['project_id'] = cluster.project_id
        extra_params['post_install_manifest_url'] = \
            CONF.kubernetes.post_install_manifest_url

        if not extra_params.get('max_node_count'):
            extra_params['max_node_count'] = cluster.node_count + 1

        self._set_cert_manager_params(context, cluster, extra_params)
        self._get_keystone_auth_default_policy(extra_params)
        self._set_volumes(context, cluster, extra_params)

        return super(K8sFedoraTemplateDefinition,
                     self).get_params(context, cluster_template, cluster,
                                      extra_params=extra_params,
                                      **kwargs)

    def _set_cert_manager_params(self, context, cluster, extra_params):
        cert_manager_api = cluster.labels.get('cert_manager_api')
        if strutils.bool_from_string(cert_manager_api):
            extra_params['cert_manager_api'] = cert_manager_api
            ca_cert = cert_manager.get_cluster_ca_certificate(cluster,
                                                              context=context)
            if isinstance(ca_cert.get_private_key_passphrase(), str):
                extra_params['ca_key'] = x509.decrypt_key(
                    ca_cert.get_private_key(),
                    ca_cert.get_private_key_passphrase().encode()
                ).decode().replace("\n", "\\n")
            else:
                extra_params['ca_key'] = x509.decrypt_key(
                    ca_cert.get_private_key(),
                    ca_cert.get_private_key_passphrase()).replace("\n", "\\n")

    def _get_keystone_auth_default_policy(self, extra_params):
        # NOTE(flwang): This purpose of this function is to make the default
        # policy more flexible for different cloud providers. Since the default
        # policy was "hardcode" in the bash script and vendors can't change
        # it unless fork it. So the new config option is introduced to address
        # this. This function can be extracted to k8s_template_def.py if k8s
        # keystone auth feature is adopted by other drivers.

        default_policy = """[{"resource": {"verbs": ["list"],
            "resources": ["pods", "services", "deployments", "pvc"],
            "version": "*", "namespace": "default"},
            "match": [{"type": "role","values": ["member"]},
            {"type": "project", "values": ["$PROJECT_ID"]}]}]"""

        keystone_auth_enabled = extra_params.get("keystone_auth_enabled",
                                                 "True")
        if strutils.bool_from_string(keystone_auth_enabled):
            try:
                with open(CONF.kubernetes.keystone_auth_default_policy) as f:
                    default_policy = json.dumps(json.loads(f.read()))
            except Exception:
                LOG.error("Failed to load default keystone auth policy")
                default_policy = json.dumps(json.loads(default_policy),
                                            sort_keys=True)

            washed_policy = default_policy.replace('"', '\"') \
                .replace("$PROJECT_ID", extra_params["project_id"])
            extra_params["keystone_auth_default_policy"] = washed_policy

    def _set_volumes(self, context, cluster, extra_params):
        # set docker_volume_type
        docker_volume_size = cluster.docker_volume_size or 0
        docker_volume_type = (cluster.labels.get(
            'docker_volume_type',
            cinder.get_default_docker_volume_type(context))
            if int(docker_volume_size) > 0 else '')
        extra_params['docker_volume_type'] = docker_volume_type

        # set etcd_volume_type
        etcd_volume_size = cluster.labels.get('etcd_volume_size', 0)
        etcd_volume_type = (cluster.labels.get(
            'etcd_volume_type',
            cinder.get_default_etcd_volume_type(context))
            if int(etcd_volume_size) > 0 else '')
        extra_params['etcd_volume_type'] = etcd_volume_type

        # set boot_volume_size
        boot_volume_size = cluster.labels.get(
            'boot_volume_size', CONF.cinder.default_boot_volume_size)
        extra_params['boot_volume_size'] = boot_volume_size

        # set boot_volume_type
        boot_volume_type = (cluster.labels.get(
            'boot_volume_type',
            cinder.get_default_boot_volume_type(context))
            if int(boot_volume_size) > 0 else '')
        extra_params['boot_volume_type'] = boot_volume_type

    def get_env_files(self, cluster_template, cluster, nodegroup=None):
        env_files = []

        template_def.add_priv_net_env_file(env_files, cluster_template,
                                           cluster)
        template_def.add_etcd_volume_env_file(env_files, cluster)
        template_def.add_volume_env_file(env_files, cluster,
                                         nodegroup=nodegroup)
        template_def.add_lb_env_file(env_files, cluster)
        template_def.add_fip_env_file(env_files, cluster)

        return env_files



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from magnum.tests.functional.api.v1.models import magnum_service_model
from magnum.tests.functional.common import client


class MagnumServiceClient(client.MagnumClient):
    """Encapsulates REST calls and maps JSON to/from models"""

    @classmethod
    def magnum_service_uri(cls, filters=None):
        """Construct magnum services uri with optional filters

        :param filters: Optional k:v dict that's converted to url query
        :returns: url string
        """

        url = "/mservices"
        if filters:
            url = cls.add_filters(url, filters)
        return url

    def magnum_service_list(self, filters=None, **kwargs):
        """Makes GET /mservices request and returns MagnumServiceCollection

        Abstracts REST call to return all magnum services.

        :param filters: Optional k:v dict that's converted to url query
        :returns: response object and MagnumServiceCollection object
        """

        resp, body = self.get(self.magnum_service_uri(filters), **kwargs)
        return self.deserialize(resp, body,
                                magnum_service_model.MagnumServiceCollection)



# Copyright (c) 2018 European Organization for Nuclear Research.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""nodegroups_v2

Revision ID: c04e925e65c2
Revises: 47380964133d
Create Date: 2019-06-14 09:29:58.288671

"""

# revision identifiers, used by Alembic.
revision = 'c04e925e65c2'
down_revision = '47380964133d'

from alembic import op  # noqa: E402

import sqlalchemy as sa  # noqa: E402

from sqlalchemy.types import String  # noqa: E402


def upgrade():
    op.add_column('nodegroup', sa.Column('stack_id', String(255)))
    op.add_column('nodegroup', sa.Column('status', String(20)))
    op.add_column('nodegroup', sa.Column('status_reason', sa.Text()))
    op.add_column('nodegroup', sa.Column('version', String(20)))

    # Populate existing nodegroups with the cluster stack_id
    connection = op.get_bind()

    connection.execute(sa.text(
        "UPDATE nodegroup "
        "INNER JOIN cluster ON nodegroup.cluster_id=cluster.uuid "
        "SET nodegroup.stack_id=cluster.stack_id, "
        "nodegroup.status=cluster.status, nodegroup.version=0 "
        "WHERE nodegroup.cluster_id=cluster.uuid")
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

conductor_group = cfg.OptGroup(name='conductor',
                               title='Options for the magnum-conductor '
                                     'service')

conductor_service_opts = [
    cfg.StrOpt('topic',
               default='magnum-conductor',
               help='The queue to add conductor tasks to.'),
    cfg.IntOpt('workers',
               help='Number of magnum-conductor processes to fork and run. '
                    'Default to number of CPUs on the host.')
]


def register_opts(conf):
    conf.register_group(conductor_group)
    conf.register_opts(conductor_service_opts, group=conductor_group)


def list_opts():
    return {
        conductor_group: conductor_service_opts
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webtest.app import AppError

from magnum.tests.unit.common.policies import base


class TestStatsPolicy(base.PolicyFunctionalTest):
    def test_stat_reader(self):
        response = self.get_json('/stats', headers=self.reader_headers)
        expected = {u'clusters': 0, u'nodes': 0}
        self.assertEqual(expected, response)

    def test_stat_admin(self):
        response = self.get_json('/stats', headers=self.admin_headers)
        expected = {u'clusters': 0, u'nodes': 0}
        self.assertEqual(expected, response)

    def test_stat_no_permission(self):
        exc = self.assertRaises(
            AppError, self.get_json, '/stats',
            headers=self.member_headers)
        self.assertIn("403 Forbidden", str(exc))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
from webtest.app import AppError

from magnum.tests.unit.api import utils as apiutils
from magnum.tests.unit.common.policies import base
from magnum.tests.unit.objects import utils as obj_utils

READER_HEADERS = {
    'OpenStack-API-Version': 'container-infra latest',
    "X-Roles": "reader"
}
HEADERS = {
    'OpenStack-API-Version': 'container-infra latest',
    "X-Roles": "member"
}


class TestCertifiactePolicy(base.PolicyFunctionalTest):
    def setUp(self):
        super(TestCertifiactePolicy, self).setUp()
        self.cluster = obj_utils.create_test_cluster(self.context)

        conductor_api_patcher = mock.patch('magnum.conductor.api.API')
        self.conductor_api_class = conductor_api_patcher.start()
        self.conductor_api = mock.MagicMock()
        self.conductor_api_class.return_value = self.conductor_api
        self.addCleanup(conductor_api_patcher.stop)

        self.conductor_api.sign_certificate.side_effect = self._fake_sign

    @staticmethod
    def _fake_sign(cluster, cert):
        cert.pem = 'fake-pem'
        return cert

    def test_get_no_permission(self):
        exc = self.assertRaises(
            AppError,
            self.get_json,
            f"/certificates/{self.cluster.uuid}",
            headers=HEADERS)
        self.assertIn("403 Forbidden", str(exc))

    def test_create_no_permission(self):
        new_cert = apiutils.cert_post_data(cluster_uuid=self.cluster.uuid)
        del new_cert['pem']

        exc = self.assertRaises(
            AppError, self.post_json,
            '/certificates', new_cert,
            headers=READER_HEADERS)
        self.assertIn("403 Forbidden", str(exc))

    def test_update_no_permission(self):
        exc = self.assertRaises(
            AppError, self.patch_json,
            f"/certificates/{self.cluster.uuid}", {},
            headers=READER_HEADERS
        )
        self.assertIn("403 Forbidden", str(exc))



# Copyright 2014 - Rackspace Hosting
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Starter script for the Magnum conductor service."""

import os
import sys

from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_reports import guru_meditation_report as gmr
from oslo_reports import opts as gmr_opts
from oslo_service import service

from magnum.common import rpc_service
from magnum.common import service as magnum_service
from magnum.common import short_id
from magnum.conductor.handlers import ca_conductor
from magnum.conductor.handlers import cluster_conductor
from magnum.conductor.handlers import conductor_listener
from magnum.conductor.handlers import federation_conductor
from magnum.conductor.handlers import indirection_api
from magnum.conductor.handlers import nodegroup_conductor
import magnum.conf
from magnum import version

CONF = magnum.conf.CONF
LOG = logging.getLogger(__name__)


def main():
    magnum_service.prepare_service(sys.argv)

    gmr_opts.set_defaults(CONF)
    gmr.TextGuruMeditation.setup_autorun(version, conf=CONF)

    LOG.info('Starting server in PID %s', os.getpid())
    LOG.debug("Configuration:")
    CONF.log_opt_values(LOG, logging.DEBUG)

    conductor_id = short_id.generate_id()
    endpoints = [
        indirection_api.Handler(),
        cluster_conductor.Handler(),
        conductor_listener.Handler(),
        ca_conductor.Handler(),
        federation_conductor.Handler(),
        nodegroup_conductor.Handler(),
    ]

    server = rpc_service.Service.create(CONF.conductor.topic,
                                        conductor_id, endpoints,
                                        binary='magnum-conductor')
    workers = CONF.conductor.workers
    if not workers:
        workers = processutils.get_worker_count()
    launcher = service.launch(CONF, server, workers=workers)

    # NOTE(mnaser): We create the periodic tasks here so that they
    #               can be attached to the main process and not
    #               duplicated in all the children if multiple
    #               workers are being used.
    server.create_periodic_tasks()
    server.start()

    launcher.wait()



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
from unittest import mock

from oslo_config import cfg
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
from wsme import types as wtypes

from magnum.api import attr_validator
from magnum.api.controllers.v1 import cluster as api_cluster
from magnum.common import exception
from magnum.conductor import api as rpcapi
import magnum.conf
from magnum import objects
from magnum.tests import base
from magnum.tests.unit.api import base as api_base
from magnum.tests.unit.api import utils as apiutils
from magnum.tests.unit.db import utils as db_utils
from magnum.tests.unit.objects import utils as obj_utils

CONF = magnum.conf.CONF


class TestClusterObject(base.TestCase):
    def test_cluster_init(self):
        cluster_dict = apiutils.cluster_post_data(cluster_template_id=None)
        del cluster_dict['node_count']
        del cluster_dict['master_count']
        del cluster_dict['create_timeout']
        cluster = api_cluster.Cluster(**cluster_dict)
        self.assertEqual(1, cluster.node_count)
        self.assertEqual(1, cluster.master_count)
        self.assertEqual(60, cluster.create_timeout)

        # test unset value for cluster_template_id
        cluster.cluster_template_id = wtypes.Unset
        self.assertEqual(wtypes.Unset, cluster.cluster_template_id)

        # test backwards compatibility of cluster fields with new objects
        cluster_dict['create_timeout'] = 15
        cluster = api_cluster.Cluster(**cluster_dict)
        self.assertEqual(15, cluster.create_timeout)


class TestListCluster(api_base.FunctionalTest):
    _cluster_attrs = ("name", "cluster_template_id", "node_count", "status",
                      "master_count", "stack_id", "create_timeout")

    _expand_cluster_attrs = ("name", "cluster_template_id", "node_count",
                             "status", "api_address", "discovery_url",
                             "node_addresses", "master_count",
                             "master_addresses", "stack_id",
                             "create_timeout", "status_reason")

    def setUp(self):
        super(TestListCluster, self).setUp()
        obj_utils.create_test_cluster_template(self.context)

    def test_empty(self):
        response = self.get_json('/clusters')
        self.assertEqual([], response['clusters'])

    def test_one(self):
        cluster = obj_utils.create_test_cluster(self.context)
        response = self.get_json('/clusters')
        self.assertEqual(cluster.uuid, response['clusters'][0]["uuid"])
        self._verify_attrs(self._cluster_attrs, response['clusters'][0])

        # Verify attrs do not appear from cluster's get_all response
        none_attrs = \
            set(self._expand_cluster_attrs) - set(self._cluster_attrs)
        self._verify_attrs(none_attrs, response['clusters'][0],
                           positive=False)

    def test_get_one(self):
        cluster = obj_utils.create_test_cluster(self.context)
        response = self.get_json('/clusters/%s' % cluster['uuid'])
        self.assertEqual(cluster.uuid, response['uuid'])
        self._verify_attrs(self._expand_cluster_attrs, response)

    def test_get_one_failed_cluster(self):
        cluster = obj_utils.create_test_cluster(self.context,
                                                status='CREATE_FAILED',
                                                master_status='CREATE_FAILED',
                                                master_reason='fake_reason')
        response = self.get_json('/clusters/%s' % cluster['uuid'])
        expected_faults = {cluster.default_ng_master.name: 'fake_reason'}
        self.assertEqual(cluster.uuid, response['uuid'])
        self.assertEqual(expected_faults, response['faults'])

    def test_get_one_by_name(self):
        cluster = obj_utils.create_test_cluster(self.context)
        response = self.get_json('/clusters/%s' % cluster['name'])
        self.assertEqual(cluster.uuid, response['uuid'])
        self._verify_attrs(self._expand_cluster_attrs, response)

    def test_get_one_by_name_not_found(self):
        response = self.get_json(
            '/clusters/not_found',
            expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_get_one_by_uuid(self):
        temp_uuid = uuidutils.generate_uuid()
        obj_utils.create_test_cluster(self.context, uuid=temp_uuid)
        response = self.get_json('/clusters/%s' % temp_uuid)
        self.assertEqual(temp_uuid, response['uuid'])
        self.assertIn('labels_overridden', response)
        self.assertIn('labels_added', response)
        self.assertIn('labels_skipped', response)

    def test_get_one_merged_labels(self):
        ct_uuid = uuidutils.generate_uuid()
        ct_labels = {'label1': 'value1', 'label2': 'value2'}
        obj_utils.create_test_cluster_template(self.context, uuid=ct_uuid,
                                               labels=ct_labels)
        c_uuid = uuidutils.generate_uuid()
        c_labels = {'label1': 'value3', 'label4': 'value4'}
        obj_utils.create_test_cluster(self.context, uuid=c_uuid,
                                      labels=c_labels,
                                      cluster_template_id=ct_uuid)
        response = self.get_json('/clusters/%s' % c_uuid)
        self.assertEqual(c_labels, response['labels'])
        self.assertEqual({'label1': 'value1'}, response['labels_overridden'])
        self.assertEqual({'label2': 'value2'}, response['labels_skipped'])
        self.assertEqual({'label4': 'value4'}, response['labels_added'])

    def test_get_one_by_uuid_not_found(self):
        temp_uuid = uuidutils.generate_uuid()
        response = self.get_json(
            '/clusters/%s' % temp_uuid,
            expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    @mock.patch("magnum.common.policy.enforce")
    @mock.patch("magnum.common.context.make_context")
    def test_get_one_by_uuid_admin(self, mock_context, mock_policy):
        temp_uuid = uuidutils.generate_uuid()
        obj_utils.create_test_cluster(self.context, uuid=temp_uuid,
                                      project_id=temp_uuid)
        self.context.is_admin = True
        response = self.get_json(
            '/clusters/%s' % temp_uuid)
        self.assertEqual(temp_uuid, response['uuid'])

    def test_get_one_by_name_multiple_cluster(self):
        obj_utils.create_test_cluster(self.context, name='test_cluster',
                                      uuid=uuidutils.generate_uuid())
        obj_utils.create_test_cluster(self.context, name='test_cluster',
                                      uuid=uuidutils.generate_uuid())
        response = self.get_json('/clusters/test_cluster',
                                 expect_errors=True)
        self.assertEqual(409, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_get_all_with_pagination_marker(self):
        cluster_list = []
        for id_ in range(4):
            temp_uuid = uuidutils.generate_uuid()
            cluster = obj_utils.create_test_cluster(self.context, id=id_,
                                                    uuid=temp_uuid)
            cluster_list.append(cluster)

        response = self.get_json('/clusters?limit=3&marker=%s'
                                 % cluster_list[2].uuid)
        self.assertEqual(1, len(response['clusters']))
        self.assertEqual(cluster_list[-1].uuid,
                         response['clusters'][0]['uuid'])

    @mock.patch("magnum.common.policy.enforce")
    @mock.patch("magnum.common.context.make_context")
    @mock.patch("magnum.objects.Cluster.obj_load_attr")
    @mock.patch("magnum.objects.Cluster.cluster_template")
    def test_get_all_with_all_projects(self, mock_context, mock_policy,
                                       mock_load, mock_template):
        for id_ in range(4):
            temp_uuid = uuidutils.generate_uuid()
            obj_utils.create_test_cluster(self.context, id=id_,
                                          uuid=temp_uuid,
                                          project_id=id_)

        self.context.is_admin = True
        response = self.get_json('/clusters')
        self.assertEqual(4, len(response['clusters']))

    def test_detail(self):
        cluster = obj_utils.create_test_cluster(self.context)
        response = self.get_json('/clusters/detail')
        self.assertEqual(cluster.uuid, response['clusters'][0]["uuid"])
        self._verify_attrs(self._expand_cluster_attrs,
                           response['clusters'][0])

    def test_detail_with_pagination_marker(self):
        cluster_list = []
        for id_ in range(4):
            temp_uuid = uuidutils.generate_uuid()
            cluster = obj_utils.create_test_cluster(self.context, id=id_,
                                                    uuid=temp_uuid)
            cluster_list.append(cluster)

        response = self.get_json('/clusters/detail?limit=3&marker=%s'
                                 % cluster_list[2].uuid)
        self.assertEqual(1, len(response['clusters']))
        self.assertEqual(cluster_list[-1].uuid,
                         response['clusters'][0]['uuid'])
        self._verify_attrs(self._expand_cluster_attrs,
                           response['clusters'][0])

    def test_detail_against_single(self):
        cluster = obj_utils.create_test_cluster(self.context)
        response = self.get_json('/clusters/%s/detail' % cluster['uuid'],
                                 expect_errors=True)
        self.assertEqual(404, response.status_int)

    def test_many(self):
        bm_list = []
        for id_ in range(5):
            temp_uuid = uuidutils.generate_uuid()
            cluster = obj_utils.create_test_cluster(self.context, id=id_,
                                                    uuid=temp_uuid)
            bm_list.append(cluster.uuid)
        response = self.get_json('/clusters')
        self.assertEqual(len(bm_list), len(response['clusters']))
        uuids = [b['uuid'] for b in response['clusters']]
        self.assertEqual(sorted(bm_list), sorted(uuids))

    def test_links(self):
        uuid = uuidutils.generate_uuid()
        obj_utils.create_test_cluster(self.context, id=1, uuid=uuid)
        response = self.get_json('/clusters/%s' % uuid)
        self.assertIn('links', response.keys())
        self.assertEqual(2, len(response['links']))
        self.assertIn(uuid, response['links'][0]['href'])
        for link in response['links']:
            bookmark = link['rel'] == 'bookmark'
            self.assertTrue(self.validate_link(link['href'],
                                               bookmark=bookmark))

    def test_collection_links(self):
        for id_ in range(5):
            obj_utils.create_test_cluster(self.context, id=id_,
                                          uuid=uuidutils.generate_uuid())
        response = self.get_json('/clusters/?limit=3')
        self.assertEqual(3, len(response['clusters']))

        next_marker = response['clusters'][-1]['uuid']
        self.assertIn(next_marker, response['next'])

    def test_collection_links_default_limit(self):
        cfg.CONF.set_override('max_limit', 3, 'api')
        for id_ in range(5):
            obj_utils.create_test_cluster(self.context, id=id_,
                                          uuid=uuidutils.generate_uuid())
        response = self.get_json('/clusters')
        self.assertEqual(3, len(response['clusters']))

        next_marker = response['clusters'][-1]['uuid']
        self.assertIn(next_marker, response['next'])


class TestPatch(api_base.FunctionalTest):
    def setUp(self):
        super(TestPatch, self).setUp()
        self.cluster_template_obj = obj_utils.create_test_cluster_template(
            self.context)
        self.cluster_obj = obj_utils.create_test_cluster(
            self.context, name='cluster_example_A', node_count=3,
            health_status='UNKNOWN', health_status_reason={})
        p = mock.patch.object(rpcapi.API, 'cluster_update_async')
        self.mock_cluster_update = p.start()
        self.mock_cluster_update.side_effect = self._sim_rpc_cluster_update
        self.addCleanup(p.stop)

    def _sim_rpc_cluster_update(self, cluster, node_count, health_status,
                                health_status_reason, rollback=False):
        cluster.status = 'UPDATE_IN_PROGRESS'
        cluster.health_status = health_status
        cluster.health_status_reason = health_status_reason
        default_ng_worker = cluster.default_ng_worker
        default_ng_worker.node_count = node_count
        default_ng_worker.save()
        cluster.save()
        return cluster

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_replace_ok(self, mock_utcnow):
        new_node_count = 4
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.patch_json('/clusters/%s' % self.cluster_obj.uuid,
                                   [{'path': '/node_count',
                                     'value': new_node_count,
                                     'op': 'replace'}])
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_code)

        response = self.get_json('/clusters/%s' % self.cluster_obj.uuid)
        self.assertEqual(new_node_count, response['node_count'])
        return_updated_at = timeutils.parse_isotime(
            response['updated_at']).replace(tzinfo=None)
        self.assertEqual(test_time, return_updated_at)
        # Assert nothing else was changed
        self.assertEqual(self.cluster_obj.uuid, response['uuid'])
        self.assertEqual(self.cluster_obj.cluster_template_id,
                         response['cluster_template_id'])

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_replace_health_status_ok(self, mock_utcnow):
        new_health_status = 'HEALTHY'
        new_health_status_reason = {'api': 'ok'}
        health_status_reason_dumps = jsonutils.dumps(new_health_status_reason)
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time
        old_node_count = self.cluster_obj.default_ng_worker.node_count
        db_utils.create_test_nodegroup(cluster_id=self.cluster_obj.uuid,
                                       name='non_default_ng')
        response = self.patch_json('/clusters/%s' % self.cluster_obj.uuid,
                                   [{'path': '/health_status',
                                     'value': new_health_status,
                                     'op': 'replace'},
                                    {'path': '/health_status_reason',
                                     'value': health_status_reason_dumps,
                                     'op': 'replace'}])
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_code)

        response = self.get_json('/clusters/%s' % self.cluster_obj.uuid)
        self.assertEqual(new_health_status, response['health_status'])
        self.assertEqual(new_health_status_reason,
                         response['health_status_reason'])

        new_node_count = self.cluster_obj.default_ng_worker.node_count
        self.assertEqual(old_node_count, new_node_count)

        return_updated_at = timeutils.parse_isotime(
            response['updated_at']).replace(tzinfo=None)
        self.assertEqual(test_time, return_updated_at)
        # Assert nothing else was changed
        self.assertEqual(self.cluster_obj.uuid, response['uuid'])
        self.assertEqual(self.cluster_obj.cluster_template_id,
                         response['cluster_template_id'])

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_replace_ok_by_name(self, mock_utcnow):
        new_node_count = 4
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.patch_json('/clusters/%s' % self.cluster_obj.name,
                                   [{'path': '/node_count',
                                     'value': new_node_count,
                                     'op': 'replace'}])
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_code)

        response = self.get_json('/clusters/%s' % self.cluster_obj.uuid)
        self.assertEqual(new_node_count, response['node_count'])
        return_updated_at = timeutils.parse_isotime(
            response['updated_at']).replace(tzinfo=None)
        self.assertEqual(test_time, return_updated_at)
        # Assert nothing else was changed
        self.assertEqual(self.cluster_obj.uuid, response['uuid'])
        self.assertEqual(self.cluster_obj.cluster_template_id,
                         response['cluster_template_id'])

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_replace_ok_by_name_not_found(self, mock_utcnow):
        name = 'not_found'
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.patch_json('/clusters/%s' % name,
                                   [{'path': '/name', 'value': name,
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(404, response.status_code)

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_replace_ok_by_uuid_not_found(self, mock_utcnow):
        uuid = uuidutils.generate_uuid()
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.patch_json('/clusters/%s' % uuid,
                                   [{'path': '/cluster_id', 'value': uuid,
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(404, response.status_code)

    def test_replace_cluster_template_id_failed(self):
        cluster_template = obj_utils.create_test_cluster_template(
            self.context,
            uuid=uuidutils.generate_uuid())
        response = self.patch_json('/clusters/%s' % self.cluster_obj.uuid,
                                   [{'path': '/cluster_template_id',
                                     'value': cluster_template.uuid,
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_code)
        self.assertTrue(response.json['errors'])

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_replace_ok_by_name_multiple_cluster(self, mock_utcnow):
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        obj_utils.create_test_cluster(self.context, name='test_cluster',
                                      uuid=uuidutils.generate_uuid())
        obj_utils.create_test_cluster(self.context, name='test_cluster',
                                      uuid=uuidutils.generate_uuid())

        response = self.patch_json('/clusters/test_cluster',
                                   [{'path': '/name',
                                     'value': 'test_cluster',
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(409, response.status_code)

    def test_replace_non_existent_cluster_template_id(self):
        response = self.patch_json('/clusters/%s' % self.cluster_obj.uuid,
                                   [{'path': '/cluster_template_id',
                                     'value': uuidutils.generate_uuid(),
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_code)
        self.assertTrue(response.json['errors'])

    def test_replace_invalid_node_count(self):
        response = self.patch_json('/clusters/%s' % self.cluster_obj.uuid,
                                   [{'path': '/node_count', 'value': -1,
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_code)
        self.assertTrue(response.json['errors'])

    def test_replace_non_existent_cluster(self):
        response = self.patch_json('/clusters/%s' %
                                   uuidutils.generate_uuid(),
                                   [{'path': '/name',
                                     'value': 'cluster_example_B',
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_replace_cluster_name_failed(self):
        response = self.patch_json('/clusters/%s' % self.cluster_obj.uuid,
                                   [{'path': '/name',
                                     'value': 'cluster_example_B',
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_add_non_existent_property(self):
        response = self.patch_json(
            '/clusters/%s' % self.cluster_obj.uuid,
            [{'path': '/foo', 'value': 'bar', 'op': 'add'}],
            expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_update_cluster_with_rollback_enabled(self):
        node_count = 4
        response = self.patch_json(
            '/clusters/%s/?rollback=True' % self.cluster_obj.uuid,
            [{'path': '/node_count', 'value': node_count,
              'op': 'replace'}],
            headers={'OpenStack-API-Version': 'container-infra 1.3',
                     "X-Roles": "member"
                     })

        self.mock_cluster_update.assert_called_once_with(
            mock.ANY, node_count, self.cluster_obj.health_status,
            self.cluster_obj.health_status_reason, True)
        self.assertEqual(202, response.status_code)

    def test_update_cluster_with_rollback_disabled(self):
        node_count = 4
        response = self.patch_json(
            '/clusters/%s/?rollback=False' % self.cluster_obj.uuid,
            [{'path': '/node_count', 'value': node_count,
              'op': 'replace'}],
            headers={'OpenStack-API-Version': 'container-infra 1.3',
                     "X-Roles": "member"
                     })

        self.mock_cluster_update.assert_called_once_with(
            mock.ANY, node_count, self.cluster_obj.health_status,
            self.cluster_obj.health_status_reason, False)
        self.assertEqual(202, response.status_code)

    def test_update_cluster_with_zero_node_count_fail(self):
        node_count = 0
        response = self.patch_json(
            '/clusters/%s' % self.cluster_obj.uuid,
            [{'path': '/node_count', 'value': node_count,
              'op': 'replace'}],
            headers={'OpenStack-API-Version': 'container-infra 1.9',
                     "X-Roles": "member"
                     },
            expect_errors=True)

        self.assertEqual(400, response.status_code)

    def test_update_cluster_with_zero_node_count(self):
        node_count = 0
        response = self.patch_json(
            '/clusters/%s' % self.cluster_obj.uuid,
            [{'path': '/node_count', 'value': node_count,
              'op': 'replace'}],
            headers={'OpenStack-API-Version': 'container-infra 1.10',
                     "X-Roles": "member"
                     })

        self.mock_cluster_update.assert_called_once_with(
            mock.ANY, node_count, self.cluster_obj.health_status,
            self.cluster_obj.health_status_reason, False)
        self.assertEqual(202, response.status_code)

    def test_remove_ok(self):
        response = self.get_json('/clusters/%s' % self.cluster_obj.uuid)
        self.assertIsNotNone(response['name'])

        response = self.patch_json('/clusters/%s' % self.cluster_obj.uuid,
                                   [{'path': '/node_count',
                                     'op': 'remove'}])
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_code)

        response = self.get_json('/clusters/%s' % self.cluster_obj.uuid)
        # only allow node_count for cluster, and default value is 1
        self.assertEqual(1, response['node_count'])
        # Assert nothing else was changed
        self.assertEqual(self.cluster_obj.uuid, response['uuid'])
        self.assertEqual(self.cluster_obj.cluster_template_id,
                         response['cluster_template_id'])
        self.assertEqual(self.cluster_obj.name, response['name'])
        self.assertEqual(self.cluster_obj.master_count,
                         response['master_count'])

    def test_remove_mandatory_property_fail(self):
        mandatory_properties = ('/uuid', '/cluster_template_id')
        for p in mandatory_properties:
            response = self.patch_json('/clusters/%s' % self.cluster_obj.uuid,
                                       [{'path': p, 'op': 'remove'}],
                                       expect_errors=True)
            self.assertEqual(400, response.status_int)
            self.assertEqual('application/json', response.content_type)
            self.assertTrue(response.json['errors'])

    def test_remove_non_existent_property(self):
        response = self.patch_json(
            '/clusters/%s' % self.cluster_obj.uuid,
            [{'path': '/non-existent', 'op': 'remove'}],
            expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_code)
        self.assertTrue(response.json['errors'])

    @mock.patch("magnum.common.policy.enforce")
    @mock.patch("magnum.common.context.make_context")
    def test_update_cluster_as_admin(self, mock_context, mock_policy):
        temp_uuid = uuidutils.generate_uuid()
        obj_utils.create_test_cluster(self.context, uuid=temp_uuid)
        self.context.is_admin = True
        response = self.patch_json('/clusters/%s' % temp_uuid,
                                   [{'path': '/node_count',
                                     'value': 4,
                                     'op': 'replace'}])
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_code)


class TestPost(api_base.FunctionalTest):
    def setUp(self):
        super(TestPost, self).setUp()
        self.cluster_template = obj_utils.create_test_cluster_template(
            self.context)
        p = mock.patch.object(rpcapi.API, 'cluster_create_async')
        self.mock_cluster_create = p.start()
        self.mock_cluster_create.side_effect = self._simulate_cluster_create
        self.addCleanup(p.stop)
        p = mock.patch.object(attr_validator, 'validate_os_resources')
        self.mock_valid_os_res = p.start()
        self.addCleanup(p.stop)

    def _simulate_cluster_create(self, cluster, master_count, node_count,
                                 create_timeout):
        cluster.create()
        return cluster

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_create_cluster(self, mock_utcnow):
        bdict = apiutils.cluster_post_data()
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        self.assertTrue(uuidutils.is_uuid_like(response.json['uuid']))

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_create_cluster_resource_limit_reached(self, mock_utcnow):
        # override max_cluster_per_project to 1
        CONF.set_override('max_clusters_per_project', 1, group='quotas')

        bdict = apiutils.cluster_post_data()
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        # create first cluster
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        self.assertTrue(uuidutils.is_uuid_like(response.json['uuid']))

        # now try to create second cluster and make sure it fails
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(403, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_cluster_set_project_id_and_user_id(self):
        bdict = apiutils.cluster_post_data()

        def _simulate_rpc_cluster_create(cluster, master_count, node_count,
                                         create_timeout):
            self.assertEqual(self.context.project_id, cluster.project_id)
            self.assertEqual(self.context.user_id, cluster.user_id)
            cluster.create()
            return cluster

        self.mock_cluster_create.side_effect = _simulate_rpc_cluster_create

        self.post_json('/clusters', bdict)

    def test_create_cluster_doesnt_contain_id(self):
        with mock.patch.object(self.dbapi, 'create_cluster',
                               wraps=self.dbapi.create_cluster) as cc_mock:
            bdict = apiutils.cluster_post_data(name='cluster_example_A')
            response = self.post_json('/clusters', bdict)
            cc_mock.assert_called_once_with(mock.ANY)
            # Check that 'id' is not in first arg of positional args
            self.assertNotIn('id', cc_mock.call_args[0][0])
            self.assertTrue(uuidutils.is_uuid_like(response.json['uuid']))

    def test_create_cluster_generate_uuid(self):
        bdict = apiutils.cluster_post_data()
        del bdict['uuid']

        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        self.assertTrue(uuidutils.is_uuid_like(response.json['uuid']))

    def test_create_cluster_no_cluster_template_id(self):
        bdict = apiutils.cluster_post_data()
        del bdict['cluster_template_id']
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)

    def test_create_cluster_with_non_existent_cluster_template_id(self):
        temp_uuid = uuidutils.generate_uuid()
        bdict = apiutils.cluster_post_data(cluster_template_id=temp_uuid)
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_cluster_with_non_existent_cluster_template_name(self):
        modelname = 'notfound'
        bdict = apiutils.cluster_post_data(cluster_template_id=modelname)
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_cluster_with_cluster_template_name(self):
        modelname = self.cluster_template.name
        bdict = apiutils.cluster_post_data(name=modelname)
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)

    def test_create_cluster_with_zero_node_count_fail(self):
        bdict = apiutils.cluster_post_data()
        bdict['node_count'] = 0
        response = self.post_json(
            '/clusters', bdict, expect_errors=True,
            headers={
                "Openstack-Api-Version": "container-infra 1.9",
                "X-Roles": "member"
            })
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)

    def test_create_cluster_with_zero_node_count(self):
        bdict = apiutils.cluster_post_data()
        bdict['node_count'] = 0
        response = self.post_json(
            '/clusters', bdict,
            headers={
                "Openstack-Api-Version": "container-infra 1.10",
                "X-Roles": "member"
            })
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)

    def test_create_cluster_with_node_count_negative(self):
        bdict = apiutils.cluster_post_data()
        bdict['node_count'] = -1
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_cluster_with_no_node_count(self):
        bdict = apiutils.cluster_post_data()
        del bdict['node_count']
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)

    def test_create_cluster_with_master_count_zero(self):
        bdict = apiutils.cluster_post_data()
        bdict['master_count'] = 0
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_cluster_with_no_master_count(self):
        bdict = apiutils.cluster_post_data()
        del bdict['master_count']
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)

    def test_create_cluster_with_even_master_count_oldmicroversion(self):
        bdict = apiutils.cluster_post_data()
        bdict['master_count'] = 2
        response = self.post_json(
            '/clusters',
            bdict,
            expect_errors=True,
            headers={"Openstack-Api-Version": "container-infra 1.9"}
        )
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_cluster_with_even_master_count(self):
        bdict = apiutils.cluster_post_data()
        bdict['master_count'] = 2
        response = self.post_json(
            '/clusters',
            bdict,
            expect_errors=True,
            headers={"Openstack-Api-Version": "container-infra 1.10"}
        )
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_cluster_with_negative_master_count(self):
        bdict = apiutils.cluster_post_data()
        bdict['master_count'] = -1
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_cluster_with_invalid_name(self):
        invalid_names = ['x' * 243, '123456', '123456test_cluster',
                         '-test_cluster', '.test_cluster', '_test_cluster', '']
        for value in invalid_names:
            bdict = apiutils.cluster_post_data(name=value)
            response = self.post_json('/clusters', bdict, expect_errors=True)
            self.assertEqual('application/json', response.content_type)
            self.assertEqual(400, response.status_int)
            self.assertTrue(response.json['errors'])

    def test_create_cluster_with_valid_name(self):
        valid_names = ['test_cluster123456', 'test-cluster', 'test.cluster',
                       'testcluster.', 'testcluster-', 'testcluster_',
                       'test.-_cluster', 'Testcluster']
        for value in valid_names:
            bdict = apiutils.cluster_post_data(name=value)
            response = self.post_json('/clusters', bdict, expect_errors=True)
            self.assertEqual('application/json', response.content_type)
            self.assertEqual(202, response.status_int)

    def test_create_cluster_without_name(self):
        bdict = apiutils.cluster_post_data()
        del bdict['name']
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)

    def test_create_cluster_with_timeout_none(self):
        bdict = apiutils.cluster_post_data()
        bdict['create_timeout'] = None
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)

    def test_create_cluster_with_no_timeout(self):
        def _simulate_rpc_cluster_create(cluster, master_count, node_count,
                                         create_timeout):
            self.assertEqual(60, create_timeout)
            cluster.create()
            return cluster

        self.mock_cluster_create.side_effect = _simulate_rpc_cluster_create
        bdict = apiutils.cluster_post_data()
        del bdict['create_timeout']
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)

    def test_create_cluster_with_timeout_negative(self):
        bdict = apiutils.cluster_post_data()
        bdict['create_timeout'] = -1
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_cluster_with_timeout_zero(self):
        bdict = apiutils.cluster_post_data()
        bdict['create_timeout'] = 0
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)

    def test_create_cluster_with_invalid_flavor(self):
        bdict = apiutils.cluster_post_data()
        self.mock_valid_os_res.side_effect = exception.FlavorNotFound(
            'test-flavor')
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(400, response.status_int)

    def test_create_cluster_with_invalid_ext_network(self):
        bdict = apiutils.cluster_post_data()
        self.mock_valid_os_res.side_effect = \
            exception.ExternalNetworkNotFound('test-net')
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(400, response.status_int)

    def test_create_cluster_with_invalid_keypair(self):
        bdict = apiutils.cluster_post_data()
        self.mock_valid_os_res.side_effect = exception.KeyPairNotFound(
            'test-key')
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(404, response.status_int)

    def test_create_cluster_with_nonexist_image(self):
        bdict = apiutils.cluster_post_data()
        self.mock_valid_os_res.side_effect = exception.ImageNotFound(
            'test-img')
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(400, response.status_int)

    def test_create_cluster_with_multi_images_same_name(self):
        bdict = apiutils.cluster_post_data()
        self.mock_valid_os_res.side_effect = exception.Conflict('test-img')
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(409, response.status_int)

    def test_create_cluster_with_no_os_distro_image(self):
        bdict = apiutils.cluster_post_data()
        self.mock_valid_os_res.side_effect = \
            exception.OSDistroFieldNotFound('img')
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(400, response.status_int)

    def test_create_cluster_with_no_lb_one_node(self):
        cluster_template = obj_utils.create_test_cluster_template(
            self.context, name='foo', uuid='foo', master_lb_enabled=False)
        bdict = apiutils.cluster_post_data(
            cluster_template_id=cluster_template.name, master_count=1)
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)

    def test_create_cluster_with_no_lb_multi_node(self):
        cluster_template = obj_utils.create_test_cluster_template(
            self.context, name='foo', uuid='foo', master_lb_enabled=False)
        bdict = apiutils.cluster_post_data(
            cluster_template_id=cluster_template.name, master_count=3,
            master_lb_enabled=False)
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)

    def test_create_cluster_with_keypair(self):
        bdict = apiutils.cluster_post_data()
        bdict['keypair'] = 'keypair2'
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        self.assertEqual('keypair2', cluster[0].keypair)

    def test_create_cluster_without_keypair(self):
        bdict = apiutils.cluster_post_data()
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        # Verify keypair from ClusterTemplate is used
        self.assertEqual('keypair1', cluster[0].keypair)

    def test_create_cluster_with_multi_keypair_same_name(self):
        bdict = apiutils.cluster_post_data()
        self.mock_valid_os_res.side_effect = exception.Conflict('keypair2')
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(409, response.status_int)

    def test_create_cluster_with_docker_volume_size(self):
        bdict = apiutils.cluster_post_data()
        bdict['docker_volume_size'] = 3
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        self.assertEqual(3, cluster[0].docker_volume_size)

    def test_create_cluster_with_labels(self):
        bdict = apiutils.cluster_post_data()
        bdict['labels'] = {'key': 'value'}
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        self.assertEqual({'key': 'value'}, cluster[0].labels)

    def test_create_cluster_without_docker_volume_size(self):
        bdict = apiutils.cluster_post_data()
        # Remove the default docker_volume_size from the cluster dict.
        del bdict['docker_volume_size']
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        # Verify docker_volume_size from ClusterTemplate is used
        self.assertEqual(20, cluster[0].docker_volume_size)

    def test_create_cluster_without_labels(self):
        bdict = apiutils.cluster_post_data()
        bdict.pop('labels')
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        # Verify labels from ClusterTemplate is used
        self.assertEqual({'key1': u'val1', 'key2': u'val2'}, cluster[0].labels)

    def test_create_cluster_with_invalid_docker_volume_size(self):
        invalid_values = [(-1, None), ('notanint', None),
                          (1, 'devicemapper'), (2, 'devicemapper')]
        for value in invalid_values:
            bdict = apiutils.cluster_post_data(docker_volume_size=value[0],
                                               docker_storage_driver=value[1])
            response = self.post_json('/clusters', bdict, expect_errors=True)
            self.assertEqual('application/json', response.content_type)
            self.assertEqual(400, response.status_int)
            self.assertTrue(response.json['errors'])

    def test_create_cluster_with_invalid_labels(self):
        bdict = apiutils.cluster_post_data(labels='invalid')
        response = self.post_json('/clusters', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_cluster_with_master_flavor_id(self):
        bdict = apiutils.cluster_post_data()
        bdict['master_flavor_id'] = 'm2.small'
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        self.assertEqual('m2.small', cluster[0].master_flavor_id)

    def test_create_cluster_without_master_flavor_id(self):
        bdict = apiutils.cluster_post_data()
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        # Verify master_flavor_id from ClusterTemplate is used
        self.assertEqual('m1.small', cluster[0].master_flavor_id)

    def test_create_cluster_with_flavor_id(self):
        bdict = apiutils.cluster_post_data()
        bdict['flavor_id'] = 'm2.small'
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        self.assertEqual('m2.small', cluster[0].flavor_id)

    def test_create_cluster_without_flavor_id(self):
        bdict = apiutils.cluster_post_data()
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        # Verify flavor_id from ClusterTemplate is used
        self.assertEqual('m1.small', cluster[0].flavor_id)

    def test_create_cluster_with_cinder_csi_disabled(self):
        self.cluster_template.volume_driver = 'cinder'
        self.cluster_template.save()
        cluster_labels = {'cinder_csi_enabled': 'false'}
        bdict = apiutils.cluster_post_data(labels=cluster_labels)
        note = 'in-tree Cinder volume driver is deprecated'
        with self.assertWarnsRegex(DeprecationWarning, note):
            response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)

    def test_create_cluster_without_merge_labels(self):
        self.cluster_template.labels = {'label1': 'value1', 'label2': 'value2'}
        self.cluster_template.save()
        cluster_labels = {'label2': 'value3', 'label4': 'value4'}
        bdict = apiutils.cluster_post_data(labels=cluster_labels)
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        self.assertEqual(cluster_labels, cluster[0].labels)

    def test_create_cluster_with_merge_labels(self):
        self.cluster_template.labels = {'label1': 'value1', 'label2': 'value2'}
        self.cluster_template.save()
        cluster_labels = {'label2': 'value3', 'label4': 'value4'}
        bdict = apiutils.cluster_post_data(labels=cluster_labels,
                                           merge_labels=True)
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        expected = self.cluster_template.labels
        expected.update(cluster_labels)
        self.assertEqual(expected, cluster[0].labels)

    def test_create_cluster_with_merge_labels_no_labels(self):
        self.cluster_template.labels = {'label1': 'value1', 'label2': 'value2'}
        self.cluster_template.save()
        bdict = apiutils.cluster_post_data(merge_labels=True)
        del bdict['labels']
        response = self.post_json('/clusters', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        cluster, timeout = self.mock_cluster_create.call_args
        self.assertEqual(self.cluster_template.labels, cluster[0].labels)


class TestDelete(api_base.FunctionalTest):
    def setUp(self):
        super(TestDelete, self).setUp()
        self.cluster_template = obj_utils.create_test_cluster_template(
            self.context)
        self.cluster = obj_utils.create_test_cluster(self.context)
        p = mock.patch.object(rpcapi.API, 'cluster_delete_async')
        self.mock_cluster_delete = p.start()
        self.mock_cluster_delete.side_effect = self._simulate_cluster_delete
        self.addCleanup(p.stop)

    def _simulate_cluster_delete(self, cluster_uuid):
        cluster = objects.Cluster.get_by_uuid(self.context, cluster_uuid)
        cluster.destroy()

    def test_delete_cluster(self):
        self.delete('/clusters/%s' % self.cluster.uuid)
        response = self.get_json('/clusters/%s' % self.cluster.uuid,
                                 expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_delete_cluster_not_found(self):
        uuid = uuidutils.generate_uuid()
        response = self.delete('/clusters/%s' % uuid, expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_delete_cluster_with_name_not_found(self):
        response = self.delete('/clusters/not_found', expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_delete_cluster_with_name(self):
        response = self.delete('/clusters/%s' % self.cluster.name,
                               expect_errors=True)
        self.assertEqual(204, response.status_int)

    def test_delete_multiple_cluster_by_name(self):
        obj_utils.create_test_cluster(self.context, name='test_cluster',
                                      uuid=uuidutils.generate_uuid())
        obj_utils.create_test_cluster(self.context, name='test_cluster',
                                      uuid=uuidutils.generate_uuid())
        response = self.delete('/clusters/test_cluster', expect_errors=True)
        self.assertEqual(409, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    @mock.patch("magnum.common.policy.enforce")
    @mock.patch("magnum.common.context.make_context")
    def test_delete_cluster_as_admin(self, mock_context, mock_policy):
        temp_uuid = uuidutils.generate_uuid()
        obj_utils.create_test_cluster(self.context, uuid=temp_uuid)
        self.context.is_admin = True
        response = self.delete('/clusters/%s' % temp_uuid,
                               expect_errors=True)
        self.assertEqual(204, response.status_int)


class TestClusterPolicyEnforcement(api_base.FunctionalTest):
    def setUp(self):
        super(TestClusterPolicyEnforcement, self).setUp()
        obj_utils.create_test_cluster_template(self.context)

    def _common_policy_check(self, rule, func, *arg, **kwarg):
        self.policy.set_rules({rule: "project:non_fake"})
        response = func(*arg, **kwarg)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_policy_disallow_get_all(self):
        self._common_policy_check(
            "cluster:get_all", self.get_json, '/clusters', expect_errors=True)

    def test_policy_disallow_get_one(self):
        self.cluster = obj_utils.create_test_cluster(self.context)
        self._common_policy_check(
            "cluster:get", self.get_json, '/clusters/%s' % self.cluster.uuid,
            expect_errors=True)

    def test_policy_disallow_detail(self):
        self._common_policy_check(
            "cluster:detail", self.get_json,
            '/clusters/%s/detail' % uuidutils.generate_uuid(),
            expect_errors=True)

    def test_policy_disallow_update(self):
        self.cluster = obj_utils.create_test_cluster(self.context,
                                                     name='cluster_example_A',
                                                     node_count=3)
        self._common_policy_check(
            "cluster:update", self.patch_json, '/clusters/%s' %
                                               self.cluster.name,
            [{'path': '/name', 'value': "new_name", 'op': 'replace'}],
            expect_errors=True)

    def test_policy_disallow_create(self):
        bdict = apiutils.cluster_post_data(name='cluster_example_A')
        self._common_policy_check(
            "cluster:create", self.post_json, '/clusters', bdict,
            expect_errors=True)

    def _simulate_cluster_delete(self, cluster_uuid):
        cluster = objects.Cluster.get_by_uuid(self.context, cluster_uuid)
        cluster.destroy()
        ngs = objects.NodeGroup.list(self.context, cluster_uuid)
        for ng in ngs:
            ng.destroy()

    def test_policy_disallow_delete(self):
        p = mock.patch.object(rpcapi.API, 'cluster_delete')
        self.mock_cluster_delete = p.start()
        self.mock_cluster_delete.side_effect = self._simulate_cluster_delete
        self.addCleanup(p.stop)
        self.cluster = obj_utils.create_test_cluster(self.context)
        self._common_policy_check(
            "cluster:delete", self.delete, '/clusters/%s' %
                                           self.cluster.uuid,
            expect_errors=True)

    def _owner_check(self, rule, func, *args, **kwargs):
        self.policy.set_rules({rule: "user_id:%(user_id)s"})
        response = func(*args, **kwargs)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_policy_only_owner_get_one(self):
        cluster = obj_utils.create_test_cluster(self.context,
                                                user_id='another')
        self._owner_check("cluster:get", self.get_json,
                          '/clusters/%s' % cluster.uuid,
                          expect_errors=True)

    def test_policy_only_owner_update(self):
        cluster = obj_utils.create_test_cluster(self.context,
                                                user_id='another')
        self._owner_check(
            "cluster:update", self.patch_json,
            '/clusters/%s' % cluster.uuid,
            [{'path': '/name', 'value': "new_name", 'op': 'replace'}],
            expect_errors=True)

    def test_policy_only_owner_delete(self):
        cluster = obj_utils.create_test_cluster(self.context,
                                                user_id='another')
        self._owner_check("cluster:delete", self.delete,
                          '/clusters/%s' % cluster.uuid,
                          expect_errors=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from magnum.i18n import _

cluster_template_group = cfg.OptGroup(name='cluster_template',
                                      title='Options for cluster_template')

cluster_template_opts = [
    cfg.ListOpt('kubernetes_allowed_network_drivers',
                default=['flannel', 'calico'],
                help=_("Allowed network drivers for kubernetes."),
                ),
    cfg.StrOpt('kubernetes_default_network_driver',
               default='flannel',
               help=_("Default network driver for kubernetes "
                      "cluster-templates."),
               ),
]


def register_opts(conf):
    conf.register_group(cluster_template_group)
    conf.register_opts(cluster_template_opts, group=cluster_template_group)


def list_opts():
    return {
        cluster_template_group: cluster_template_opts
    }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
from unittest.mock import patch

from heatclient import exc as heatexc
from oslo_utils import uuidutils

import magnum.conf
from magnum.drivers.heat import driver as heat_driver
from magnum.drivers.k8s_fedora_coreos_v1 import driver as k8s_fcos_dr
from magnum import objects
from magnum.objects.fields import ClusterStatus as cluster_status
from magnum.tests import base
from magnum.tests.unit.db import utils

CONF = magnum.conf.CONF


class TestHeatPoller(base.TestCase):

    def setUp(self):
        super(TestHeatPoller, self).setUp()
        self.mock_stacks = dict()
        self.def_ngs = list()

    def _create_nodegroup(self, cluster, uuid, stack_id, name=None, role=None,
                          is_default=False, stack_status=None,
                          status_reason=None, stack_params=None,
                          stack_missing=False):
        """Create a new nodegroup

        Util that creates a new non-default ng, adds it to the cluster
        and creates the corresponding mock stack.
        """
        role = 'worker' if role is None else role
        ng = mock.MagicMock(uuid=uuid, role=role, is_default=is_default,
                            stack_id=stack_id)
        if name is not None:
            type(ng).name = name

        cluster.nodegroups.append(ng)

        if stack_status is None:
            stack_status = cluster_status.CREATE_COMPLETE

        if status_reason is None:
            status_reason = 'stack created'

        stack_params = dict() if stack_params is None else stack_params

        stack = mock.MagicMock(stack_status=stack_status,
                               stack_status_reason=status_reason,
                               parameters=stack_params)
        # In order to simulate a stack not found from osc we don't add the
        # stack in the dict.
        if not stack_missing:
            self.mock_stacks.update({stack_id: stack})
        else:
            # In case the stack is missing we need
            # to set the status to the ng, so that
            # _sync_missing_heat_stack knows which
            # was the previous state.
            ng.status = stack_status

        return ng

    @patch('magnum.conductor.utils.retrieve_cluster_template')
    @patch('oslo_config.cfg')
    @patch('magnum.common.clients.OpenStackClients')
    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def setup_poll_test(self, mock_driver, mock_openstack_client, cfg,
                        mock_retrieve_cluster_template,
                        default_stack_status=None, status_reason=None,
                        stack_params=None, stack_missing=False):
        cfg.CONF.cluster_heat.max_attempts = 10

        if default_stack_status is None:
            default_stack_status = cluster_status.CREATE_COMPLETE

        cluster = mock.MagicMock(nodegroups=list(),
                                 uuid=uuidutils.generate_uuid())

        def_worker = self._create_nodegroup(cluster, 'worker_ng', 'stack1',
                                            name='worker_ng', role='worker',
                                            is_default=True,
                                            stack_status=default_stack_status,
                                            status_reason=status_reason,
                                            stack_params=stack_params,
                                            stack_missing=stack_missing)
        def_master = self._create_nodegroup(cluster, 'master_ng', 'stack1',
                                            name='master_ng', role='master',
                                            is_default=True,
                                            stack_status=default_stack_status,
                                            status_reason=status_reason,
                                            stack_params=stack_params,
                                            stack_missing=stack_missing)

        cluster.default_ng_worker = def_worker
        cluster.default_ng_master = def_master

        self.def_ngs = [def_worker, def_master]

        def get_ng_stack(stack_id, resolve_outputs=False):
            try:
                return self.mock_stacks[stack_id]
            except KeyError:
                # In this case we intentionally didn't add the stack
                # to the mock_stacks dict to simulte a not found error.
                # For this reason raise heat NotFound exception.
                raise heatexc.NotFound("stack not found")

        cluster_template_dict = utils.get_test_cluster_template(
            coe='kubernetes')
        mock_heat_client = mock.MagicMock()
        mock_heat_client.stacks.get = get_ng_stack
        mock_openstack_client.heat.return_value = mock_heat_client
        cluster_template = objects.ClusterTemplate(self.context,
                                                   **cluster_template_dict)
        mock_retrieve_cluster_template.return_value = cluster_template
        mock_driver.return_value = k8s_fcos_dr.Driver()
        poller = heat_driver.HeatPoller(mock_openstack_client,
                                        mock.MagicMock(), cluster,
                                        k8s_fcos_dr.Driver())
        poller.get_version_info = mock.MagicMock()
        return (cluster, poller)

    def test_poll_and_check_creating(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.CREATE_IN_PROGRESS)

        cluster.status = cluster_status.CREATE_IN_PROGRESS
        poller.poll_and_check()

        for ng in cluster.nodegroups:
            self.assertEqual(cluster_status.CREATE_IN_PROGRESS, ng.status)

        self.assertEqual(cluster_status.CREATE_IN_PROGRESS, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_create_complete(self):
        cluster, poller = self.setup_poll_test()

        cluster.status = cluster_status.CREATE_IN_PROGRESS
        poller.poll_and_check()

        for ng in cluster.nodegroups:
            self.assertEqual(cluster_status.CREATE_COMPLETE, ng.status)
            self.assertEqual('stack created', ng.status_reason)
            self.assertEqual(1, ng.save.call_count)

        self.assertEqual(cluster_status.CREATE_COMPLETE, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_create_failed(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.CREATE_FAILED)

        cluster.status = cluster_status.CREATE_IN_PROGRESS
        self.assertIsNone(poller.poll_and_check())

        for ng in cluster.nodegroups:
            self.assertEqual(cluster_status.CREATE_FAILED, ng.status)
            # Two calls to save since the stack ouptputs are synced too.
            self.assertEqual(2, ng.save.call_count)

        self.assertEqual(cluster_status.CREATE_FAILED, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_updating(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.UPDATE_IN_PROGRESS)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for ng in cluster.nodegroups:
            self.assertEqual(cluster_status.UPDATE_IN_PROGRESS, ng.status)
            self.assertEqual(1, ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_IN_PROGRESS, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_update_complete(self):
        stack_params = {
            'number_of_minions': 2,
            'number_of_masters': 1
        }
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.UPDATE_COMPLETE,
            stack_params=stack_params)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        self.assertIsNone(poller.poll_and_check())

        for ng in cluster.nodegroups:
            self.assertEqual(cluster_status.UPDATE_COMPLETE, ng.status)

        self.assertEqual(2, cluster.default_ng_worker.save.call_count)
        self.assertEqual(2, cluster.default_ng_master.save.call_count)
        self.assertEqual(2, cluster.default_ng_worker.node_count)
        self.assertEqual(1, cluster.default_ng_master.node_count)

        self.assertEqual(cluster_status.UPDATE_COMPLETE, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_update_failed(self):
        stack_params = {
            'number_of_minions': 2,
            'number_of_masters': 1
        }
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.UPDATE_FAILED,
            stack_params=stack_params)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for ng in cluster.nodegroups:
            self.assertEqual(cluster_status.UPDATE_FAILED, ng.status)
            # We have several calls to save because the stack outputs are
            # stored too.
            self.assertEqual(3, ng.save.call_count)

        self.assertEqual(2, cluster.default_ng_worker.node_count)
        self.assertEqual(1, cluster.default_ng_master.node_count)

        self.assertEqual(cluster_status.UPDATE_FAILED, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_deleting(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.DELETE_IN_PROGRESS)

        cluster.status = cluster_status.DELETE_IN_PROGRESS
        poller.poll_and_check()

        for ng in cluster.nodegroups:
            self.assertEqual(cluster_status.DELETE_IN_PROGRESS, ng.status)
            # We have two calls to save because the stack outputs are
            # stored too.
            self.assertEqual(1, ng.save.call_count)

        self.assertEqual(cluster_status.DELETE_IN_PROGRESS, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_deleted(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.DELETE_COMPLETE)

        cluster.status = cluster_status.DELETE_IN_PROGRESS
        self.assertIsNone(poller.poll_and_check())

        self.assertEqual(cluster_status.DELETE_COMPLETE,
                         cluster.default_ng_worker.status)
        self.assertEqual(1, cluster.default_ng_worker.save.call_count)
        self.assertEqual(0, cluster.default_ng_worker.destroy.call_count)

        self.assertEqual(cluster_status.DELETE_COMPLETE,
                         cluster.default_ng_master.status)
        self.assertEqual(1, cluster.default_ng_master.save.call_count)
        self.assertEqual(0, cluster.default_ng_master.destroy.call_count)

        self.assertEqual(cluster_status.DELETE_COMPLETE, cluster.status)
        self.assertEqual(1, cluster.save.call_count)
        self.assertEqual(0, cluster.destroy.call_count)

    def test_poll_and_check_delete_failed(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.DELETE_FAILED)

        cluster.status = cluster_status.DELETE_IN_PROGRESS
        poller.poll_and_check()

        self.assertEqual(cluster_status.DELETE_FAILED,
                         cluster.default_ng_worker.status)
        # We have two calls to save because the stack outputs are
        # stored too.
        self.assertEqual(2, cluster.default_ng_worker.save.call_count)
        self.assertEqual(0, cluster.default_ng_worker.destroy.call_count)

        self.assertEqual(cluster_status.DELETE_FAILED,
                         cluster.default_ng_master.status)
        # We have two calls to save because the stack outputs are
        # stored too.
        self.assertEqual(2, cluster.default_ng_master.save.call_count)
        self.assertEqual(0, cluster.default_ng_master.destroy.call_count)

        self.assertEqual(cluster_status.DELETE_FAILED, cluster.status)
        self.assertEqual(1, cluster.save.call_count)
        self.assertEqual(0, cluster.destroy.call_count)

    def test_poll_done_rollback_complete(self):
        stack_params = {
            'number_of_minions': 1,
            'number_of_masters': 1
        }
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.ROLLBACK_COMPLETE,
            stack_params=stack_params)

        self.assertIsNone(poller.poll_and_check())

        self.assertEqual(1, cluster.save.call_count)
        self.assertEqual(cluster_status.ROLLBACK_COMPLETE, cluster.status)
        self.assertEqual(1, cluster.default_ng_worker.node_count)
        self.assertEqual(1, cluster.default_ng_master.node_count)

    def test_poll_done_rollback_failed(self):
        stack_params = {
            'number_of_minions': 1,
            'number_of_masters': 1
        }
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.ROLLBACK_FAILED,
            stack_params=stack_params)

        self.assertIsNone(poller.poll_and_check())

        self.assertEqual(1, cluster.save.call_count)
        self.assertEqual(cluster_status.ROLLBACK_FAILED, cluster.status)
        self.assertEqual(1, cluster.default_ng_worker.node_count)
        self.assertEqual(1, cluster.default_ng_master.node_count)

    def test_poll_and_check_new_ng_creating(self):
        cluster, poller = self.setup_poll_test()

        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.CREATE_IN_PROGRESS)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.CREATE_IN_PROGRESS, ng.status)
        self.assertEqual(1, ng.save.call_count)
        self.assertEqual(cluster_status.UPDATE_IN_PROGRESS, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_new_ng_created(self):
        cluster, poller = self.setup_poll_test()

        ng = self._create_nodegroup(cluster, 'ng1', 'stack2')

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.CREATE_COMPLETE, ng.status)
        self.assertEqual(1, ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_COMPLETE, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_new_ng_create_failed(self):
        cluster, poller = self.setup_poll_test()

        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.CREATE_FAILED,
            status_reason='stack failed')

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual('stack created', def_ng.status_reason)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.CREATE_FAILED, ng.status)
        self.assertEqual('stack failed', ng.status_reason)
        self.assertEqual(2, ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_FAILED, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_new_ng_updated(self):
        cluster, poller = self.setup_poll_test()

        stack_params = {'number_of_minions': 3}
        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.UPDATE_COMPLETE,
            stack_params=stack_params)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_COMPLETE, ng.status)
        self.assertEqual(3, ng.node_count)
        self.assertEqual(2, ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_COMPLETE, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_new_ng_update_failed(self):
        cluster, poller = self.setup_poll_test()

        stack_params = {'number_of_minions': 3}
        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.UPDATE_FAILED,
            stack_params=stack_params)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_FAILED, ng.status)
        self.assertEqual(3, ng.node_count)
        self.assertEqual(3, ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_FAILED, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_new_ng_deleting(self):
        cluster, poller = self.setup_poll_test()

        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.DELETE_IN_PROGRESS)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.DELETE_IN_PROGRESS, ng.status)
        self.assertEqual(1, ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_IN_PROGRESS, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_new_ng_deleted(self):
        cluster, poller = self.setup_poll_test()

        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.DELETE_COMPLETE)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(1, ng.destroy.call_count)

        self.assertEqual(cluster_status.UPDATE_COMPLETE, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_new_ng_delete_failed(self):
        cluster, poller = self.setup_poll_test()

        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.DELETE_FAILED)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.DELETE_FAILED, ng.status)
        self.assertEqual(2, ng.save.call_count)
        self.assertEqual(0, ng.destroy.call_count)

        self.assertEqual(cluster_status.UPDATE_FAILED, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_new_ng_rollback_complete(self):
        cluster, poller = self.setup_poll_test()

        stack_params = {
            'number_of_minions': 2,
            'number_of_masters': 0
        }
        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.ROLLBACK_COMPLETE,
            stack_params=stack_params)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.ROLLBACK_COMPLETE, ng.status)
        self.assertEqual(2, ng.node_count)
        self.assertEqual(3, ng.save.call_count)
        self.assertEqual(0, ng.destroy.call_count)

        self.assertEqual(cluster_status.UPDATE_COMPLETE, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_new_ng_rollback_failed(self):
        cluster, poller = self.setup_poll_test()

        stack_params = {
            'number_of_minions': 2,
            'number_of_masters': 0
        }
        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.ROLLBACK_FAILED,
            stack_params=stack_params)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.ROLLBACK_FAILED, ng.status)
        self.assertEqual(2, ng.node_count)
        self.assertEqual(3, ng.save.call_count)
        self.assertEqual(0, ng.destroy.call_count)

        self.assertEqual(cluster_status.UPDATE_FAILED, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_multiple_new_ngs(self):
        cluster, poller = self.setup_poll_test()

        ng1 = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.CREATE_COMPLETE)
        ng2 = self._create_nodegroup(
            cluster, 'ng2', 'stack3',
            stack_status=cluster_status.UPDATE_IN_PROGRESS)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.CREATE_COMPLETE, ng1.status)
        self.assertEqual(1, ng1.save.call_count)
        self.assertEqual(cluster_status.UPDATE_IN_PROGRESS, ng2.status)
        self.assertEqual(1, ng2.save.call_count)

        self.assertEqual(cluster_status.UPDATE_IN_PROGRESS, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_multiple_ngs_failed_and_updating(self):
        cluster, poller = self.setup_poll_test()

        ng1 = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.CREATE_FAILED)
        ng2 = self._create_nodegroup(
            cluster, 'ng2', 'stack3',
            stack_status=cluster_status.UPDATE_IN_PROGRESS)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
            self.assertEqual(1, def_ng.save.call_count)

        self.assertEqual(cluster_status.CREATE_FAILED, ng1.status)
        self.assertEqual(2, ng1.save.call_count)
        self.assertEqual(cluster_status.UPDATE_IN_PROGRESS, ng2.status)
        self.assertEqual(1, ng2.save.call_count)

        self.assertEqual(cluster_status.UPDATE_IN_PROGRESS, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    @patch('magnum.drivers.heat.driver.trust_manager')
    @patch('magnum.drivers.heat.driver.cert_manager')
    def test_delete_complete(self, cert_manager, trust_manager):
        cluster, poller = self.setup_poll_test()
        poller._delete_complete()
        self.assertEqual(
            1, cert_manager.delete_certificates_from_cluster.call_count)
        self.assertEqual(1, trust_manager.delete_trustee_and_trust.call_count)

    @patch('magnum.drivers.heat.driver.LOG')
    def test_nodegroup_failed(self, logger):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.CREATE_FAILED)

        self._create_nodegroup(cluster, 'ng1', 'stack2',
                               stack_status=cluster_status.CREATE_FAILED)
        poller.poll_and_check()
        # Verify that we have one log for each failed nodegroup
        self.assertEqual(3, logger.error.call_count)

    def test_stack_not_found_creating(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.CREATE_IN_PROGRESS,
            stack_missing=True)
        poller.poll_and_check()
        for ng in cluster.nodegroups:
            self.assertEqual(cluster_status.CREATE_FAILED, ng.status)

    def test_stack_not_found_updating(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.UPDATE_IN_PROGRESS,
            stack_missing=True)
        poller.poll_and_check()
        for ng in cluster.nodegroups:
            self.assertEqual(cluster_status.UPDATE_FAILED, ng.status)

    def test_stack_not_found_deleting(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.DELETE_IN_PROGRESS,
            stack_missing=True)
        poller.poll_and_check()
        for ng in cluster.nodegroups:
            self.assertEqual(cluster_status.DELETE_COMPLETE, ng.status)

    def test_stack_not_found_new_ng_creating(self):
        cluster, poller = self.setup_poll_test()
        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.CREATE_IN_PROGRESS, stack_missing=True)
        poller.poll_and_check()
        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
        self.assertEqual(cluster_status.CREATE_FAILED, ng.status)

    def test_stack_not_found_new_ng_updating(self):
        cluster, poller = self.setup_poll_test()
        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.UPDATE_IN_PROGRESS, stack_missing=True)
        poller.poll_and_check()
        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
        self.assertEqual(cluster_status.UPDATE_FAILED, ng.status)

    def test_stack_not_found_new_ng_deleting(self):
        cluster, poller = self.setup_poll_test()
        ng = self._create_nodegroup(
            cluster, 'ng1', 'stack2',
            stack_status=cluster_status.DELETE_IN_PROGRESS, stack_missing=True)
        poller.poll_and_check()
        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.CREATE_COMPLETE, def_ng.status)
        self.assertEqual(cluster_status.DELETE_COMPLETE, ng.status)

    def test_poll_and_check_failed_default_ng(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.UPDATE_FAILED)

        ng = self._create_nodegroup(
            cluster, 'ng', 'stack2',
            stack_status=cluster_status.UPDATE_COMPLETE)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.UPDATE_FAILED, def_ng.status)
            self.assertEqual(2, def_ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_COMPLETE, ng.status)
        self.assertEqual(1, ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_FAILED, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_rollback_failed_default_ng(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.ROLLBACK_FAILED)

        ng = self._create_nodegroup(
            cluster, 'ng', 'stack2',
            stack_status=cluster_status.UPDATE_COMPLETE)

        cluster.status = cluster_status.UPDATE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.ROLLBACK_FAILED, def_ng.status)
            self.assertEqual(2, def_ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_COMPLETE, ng.status)
        self.assertEqual(1, ng.save.call_count)

        self.assertEqual(cluster_status.UPDATE_FAILED, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_rollback_failed_def_ng(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.DELETE_FAILED)

        ng = self._create_nodegroup(
            cluster, 'ng', 'stack2',
            stack_status=cluster_status.DELETE_IN_PROGRESS)

        cluster.status = cluster_status.DELETE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.DELETE_FAILED, def_ng.status)
            self.assertEqual(2, def_ng.save.call_count)

        self.assertEqual(cluster_status.DELETE_IN_PROGRESS, ng.status)
        self.assertEqual(1, ng.save.call_count)

        self.assertEqual(cluster_status.DELETE_IN_PROGRESS, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

    def test_poll_and_check_delete_failed_def_ng(self):
        cluster, poller = self.setup_poll_test(
            default_stack_status=cluster_status.DELETE_FAILED)

        ng = self._create_nodegroup(
            cluster, 'ng', 'stack2',
            stack_status=cluster_status.DELETE_COMPLETE)

        cluster.status = cluster_status.DELETE_IN_PROGRESS
        poller.poll_and_check()

        for def_ng in self.def_ngs:
            self.assertEqual(cluster_status.DELETE_FAILED, def_ng.status)
            self.assertEqual(2, def_ng.save.call_count)

        # Check that the non-default ng was deleted
        self.assertEqual(1, ng.destroy.call_count)

        self.assertEqual(cluster_status.DELETE_FAILED, cluster.status)
        self.assertEqual(1, cluster.save.call_count)

        self.assertIn('worker_ng', cluster.status_reason)
        self.assertIn('master_ng', cluster.status_reason)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utility for fetching a resource (e.g. a manifest) from a URL."""
import urllib

from oslo_log import log as logging
import requests
from requests import exceptions

from magnum.common import exception
import magnum.conf
from magnum.i18n import _

CONF = magnum.conf.CONF
LOG = logging.getLogger(__name__)


class URLFetchError(exception.Invalid, IOError):
    pass


def get(url, allowed_schemes=('http', 'https')):
    """Get the data at the specified URL.

    The URL must use the http: or https: schemes.
    Raise an IOError if getting the data fails.
    """
    LOG.info('Fetching data from %s', url)

    components = urllib.parse.urlparse(url)

    if components.scheme not in allowed_schemes:
        raise URLFetchError(_('Invalid URL scheme %s') % components.scheme)

    try:
        resp = requests.get(url, stream=True, timeout=60)
        resp.raise_for_status()

        # We cannot use resp.text here because it would download the
        # entire file, and a large enough file would bring down the
        # engine.  The 'Content-Length' header could be faked, so it's
        # necessary to download the content in chunks to until
        # max_manifest_size is reached.  The chunk_size we use needs
        # to balance CPU-intensive string concatenation with accuracy
        # (eg. it's possible to fetch 1000 bytes greater than
        # max_manifest_size with a chunk_size of 1000).
        reader = resp.iter_content(chunk_size=1000)
        result = ""
        for chunk in reader:
            result += chunk
            if len(result) > CONF.max_manifest_size:
                raise URLFetchError("Manifest exceeds maximum allowed "
                                    "size (%s bytes)" %
                                    CONF.max_manifest_size)
        return result

    except exceptions.RequestException as ex:
        raise URLFetchError(_('Failed to retrieve manifest: %s') % ex)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import os

_TRUE_VALUES = ('True', 'true', '1', 'yes')


class OutputStreamCapture(fixtures.Fixture):
    """Capture output streams during tests.

    This fixture captures errant printing to stderr / stdout during
    the tests and lets us see those streams at the end of the test
    runs instead. Useful to see what was happening during failed
    tests.
    """
    def setUp(self):
        super(OutputStreamCapture, self).setUp()
        if os.environ.get('OS_STDOUT_CAPTURE') in _TRUE_VALUES:
            self.out = self.useFixture(fixtures.StringStream('stdout'))
            self.useFixture(
                fixtures.MonkeyPatch('sys.stdout', self.out.stream))
        if os.environ.get('OS_STDERR_CAPTURE') in _TRUE_VALUES:
            self.err = self.useFixture(fixtures.StringStream('stderr'))
            self.useFixture(
                fixtures.MonkeyPatch('sys.stderr', self.err.stream))

    @property
    def stderr(self):
        return self.err._details["stderr"].as_text()

    @property
    def stdout(self):
        return self.out._details["stdout"].as_text()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from magnum.common import exception
from magnum.common import utils

LOG = logging.getLogger(__name__)


def create_trustee_and_trust(osc, cluster):
    try:
        password = utils.generate_password(length=18)

        trustee = osc.keystone().create_trustee(
            "%s_%s" % (cluster.uuid, cluster.project_id),
            password,
        )

        cluster.trustee_username = trustee.name
        cluster.trustee_user_id = trustee.id
        cluster.trustee_password = password

        trust = osc.keystone().create_trust(
            cluster.trustee_user_id)
        cluster.trust_id = trust.id

    except Exception:
        LOG.exception(
            'Failed to create trustee and trust for Cluster: %s',
            cluster.uuid)
        raise exception.TrusteeOrTrustToClusterFailed(
            cluster_uuid=cluster.uuid)


def delete_trustee_and_trust(osc, context, cluster):
    kst = osc.keystone()
    try:
        if cluster.trust_id:
            kst.delete_trust(context, cluster)
            cluster.trust_id = None
    except Exception:
        # Exceptions are already logged by keystone().delete_trust
        pass
    try:
        if cluster.trustee_user_id:
            kst.delete_trustee(cluster.trustee_user_id)
            cluster.trustee_user_id = None
            cluster.trustee_username = None
            cluster.trustee_password = None
    except Exception:
        # Exceptions are already logged by keystone().delete_trustee
        pass



# Copyright 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Magnum object test utilities."""


import datetime

import netaddr
from oslo_utils import timeutils

from magnum.common import exception
from magnum.i18n import _
from magnum import objects
from magnum.tests.unit.db import utils as db_utils


def get_test_cluster_template(context, **kw):
    """Return a ClusterTemplate object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_cluster_template = db_utils.get_test_cluster_template(**kw)
    cluster_template = objects.ClusterTemplate(context)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_cluster_template['id']

    for key in db_cluster_template:
        setattr(cluster_template, key, db_cluster_template[key])
    return cluster_template


def create_test_cluster_template(context, **kw):
    """Create and return a test ClusterTemplate object.

    Create a ClusterTemplate in the DB and return a ClusterTemplate object
    with appropriate attributes.
    """
    cluster_template = get_test_cluster_template(context, **kw)
    try:
        cluster_template.create()
    except exception.ClusterTemplateAlreadyExists:
        cluster_template = objects.ClusterTemplate.get(context,
                                                       cluster_template.uuid)
    return cluster_template


def get_test_cluster(context, **kw):
    """Return a Cluster object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_cluster = db_utils.get_test_cluster(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_cluster['id']
    cluster = objects.Cluster(context)
    for key in db_cluster:
        setattr(cluster, key, db_cluster[key])
    return cluster


def create_test_cluster(context, **kw):
    """Create and return a test Cluster object.

    Create a Cluster in the DB and return a Cluster object with appropriate
    attributes.
    """
    cluster = get_test_cluster(context, **kw)
    create_test_cluster_template(context, uuid=cluster['cluster_template_id'],
                                 coe=kw.get('coe', 'kubernetes'),
                                 tls_disabled=kw.get('tls_disabled'))
    kw.update({'cluster_id': cluster['uuid']})
    db_utils.create_nodegroups_for_cluster(**kw)
    cluster.create()
    return cluster


def get_test_quota(context, **kw):
    """Return a Quota object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_quota = db_utils.get_test_quota(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_quota['id']
    quota = objects.Quota(context)
    for key in db_quota:
        setattr(quota, key, db_quota[key])
    return quota


def create_test_quota(context, **kw):
    """Create and return a test Quota object.

    Create a quota in the DB and return a Quota object with appropriate
    attributes.
    """
    quota = get_test_quota(context, **kw)
    quota.create()
    return quota


def get_test_x509keypair(context, **kw):
    """Return a X509KeyPair object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_x509keypair = db_utils.get_test_x509keypair(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_x509keypair['id']
    x509keypair = objects.X509KeyPair(context)
    for key in db_x509keypair:
        setattr(x509keypair, key, db_x509keypair[key])
    return x509keypair


def create_test_x509keypair(context, **kw):
    """Create and return a test x509keypair object.

    Create a x509keypair in the DB and return a X509KeyPair object with
    appropriate attributes.
    """
    x509keypair = get_test_x509keypair(context, **kw)
    x509keypair.create()
    return x509keypair


def get_test_magnum_service_object(context, **kw):
    """Return a test magnum_service object.

    Get a magnum_service from DB layer and return an object with
    appropriate attributes.
    """
    db_magnum_service = db_utils.get_test_magnum_service(**kw)
    magnum_service = objects.MagnumService(context)
    for key in db_magnum_service:
        setattr(magnum_service, key, db_magnum_service[key])
    return magnum_service


def get_test_nodegroup(context, **kw):
    db_nodegroup = db_utils.get_test_nodegroup(**kw)
    nodegroup = objects.NodeGroup(context)
    for key in db_nodegroup:
        setattr(nodegroup, key, db_nodegroup[key])
    return nodegroup


def create_test_nodegroup(context, **kw):
    nodegroup = get_test_nodegroup(context, **kw)
    nodegroup.create()
    return nodegroup


def get_test_federation(context, **kw):
    """Return a Federation object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_federation = db_utils.get_test_federation(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_federation['id']
    federation = objects.Federation(context)
    for key in db_federation:
        setattr(federation, key, db_federation[key])
    return federation


def create_test_federation(context, **kw):
    """Create and return a test Federation object.

    Create a Federation in the DB and return a Federation object with
    appropriate attributes.
    """
    federation = get_test_federation(context, **kw)
    federation.create()
    return federation


def datetime_or_none(dt):
    """Validate a datetime or None value."""
    if dt is None:
        return None
    elif isinstance(dt, datetime.datetime):
        if dt.utcoffset() is None:
            # NOTE(danms): Legacy objects from sqlalchemy are stored in UTC,
            # but are returned without a timezone attached.
            # As a transitional aid, assume a tz-naive object is in UTC.
            return dt.replace(tzinfo=datetime.timezone.utc)
        else:
            return dt
    raise ValueError(_("A datetime.datetime is required here"))


def datetime_or_str_or_none(val):
    if isinstance(val, str):
        return timeutils.parse_isotime(val)
    return datetime_or_none(val)


def int_or_none(val):
    """Attempt to parse an integer value, or None."""
    if val is None:
        return val
    else:
        return int(val)


def str_or_none(val):
    """Attempt to stringify a value to unicode, or None."""
    if val is None:
        return val
    else:
        return str(val)


def ip_or_none(version):
    """Return a version-specific IP address validator."""
    def validator(val, version=version):
        if val is None:
            return val
        else:
            return netaddr.IPAddress(val, version=version)
    return validator


def dt_serializer(name):
    """Return a datetime serializer for a named attribute."""
    def serializer(self, name=name):
        if getattr(self, name) is not None:
            return datetime.datetime.isoformat(getattr(self, name))
        else:
            return None
    return serializer


def dt_deserializer(instance, val):
    """A deserializer method for datetime attributes."""
    if val is None:
        return None
    else:
        return timeutils.parse_isotime(val)



#    Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Magnum common internal object model"""

from oslo_versionedobjects import base as ovoo_base
from oslo_versionedobjects import fields as ovoo_fields


remotable_classmethod = ovoo_base.remotable_classmethod
remotable = ovoo_base.remotable


class MagnumObjectRegistry(ovoo_base.VersionedObjectRegistry):
    pass


class MagnumObject(ovoo_base.VersionedObject):
    """Base class and object factory.

    This forms the base of all objects that can be remoted or instantiated
    via RPC. Simply defining a class that inherits from this base class
    will make it remotely instantiatable. Objects should implement the
    necessary "get" classmethod routines as well as "save" object methods
    as appropriate.
    """

    OBJ_SERIAL_NAMESPACE = 'magnum_object'
    OBJ_PROJECT_NAMESPACE = 'magnum'

    def as_dict(self):
        return {k: getattr(self, k)
                for k in self.fields
                if self.obj_attr_is_set(k)}


class MagnumObjectDictCompat(ovoo_base.VersionedObjectDictCompat):
    pass


class MagnumPersistentObject(object):
    """Mixin class for Persistent objects.

    This adds the fields that we use in common for all persistent objects.
    """
    fields = {
        'created_at': ovoo_fields.DateTimeField(nullable=True),
        'updated_at': ovoo_fields.DateTimeField(nullable=True),
    }


class MagnumObjectIndirectionAPI(ovoo_base.VersionedObjectIndirectionAPI):
    def __init__(self):
        super(MagnumObjectIndirectionAPI, self).__init__()
        from magnum.conductor import api as conductor_api
        self._conductor = conductor_api.API()

    def object_action(self, context, objinst, objmethod, args, kwargs):
        return self._conductor.object_action(context, objinst, objmethod,
                                             args, kwargs)

    def object_class_action(self, context, objname, objmethod, objver,
                            args, kwargs):
        return self._conductor.object_class_action(context, objname, objmethod,
                                                   objver, args, kwargs)

    def object_backport(self, context, objinst, target_version):
        return self._conductor.object_backport(context, objinst,
                                               target_version)


class MagnumObjectSerializer(ovoo_base.VersionedObjectSerializer):
    # Base class to use for object hydration
    OBJ_BASE_CLASS = MagnumObject



# Copyright 2016 NEC Corporation.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
from unittest.mock import patch

from magnum.common import exception
from magnum.conductor.handlers.common import trust_manager
from magnum.tests import base


class TrustManagerTestCase(base.BaseTestCase):
    def setUp(self):
        super(TrustManagerTestCase, self).setUp()

        osc_class_patcher = patch('magnum.common.clients.OpenStackClients')
        osc_class = osc_class_patcher.start()
        self.addCleanup(osc_class_patcher.stop)

        self.osc = mock.MagicMock()
        osc_class.return_value = self.osc

    @patch('magnum.common.utils.generate_password')
    def test_create_trustee_and_trust(self, mock_generate_password):
        mock_password = "password_mock"
        mock_generate_password.return_value = mock_password
        mock_cluster = mock.MagicMock()
        mock_cluster.uuid = 'mock_cluster_uuid'
        mock_cluster.project_id = 'mock_cluster_project_id'
        mock_keystone = mock.MagicMock()
        mock_trustee = mock.MagicMock()
        mock_trustee.id = 'mock_trustee_id'
        mock_trustee.name = 'mock_trustee_username'
        mock_trust = mock.MagicMock()
        mock_trust.id = 'mock_trust_id'

        self.osc.keystone.return_value = mock_keystone

        mock_keystone.create_trustee.return_value = mock_trustee
        mock_keystone.create_trust.return_value = mock_trust

        trust_manager.create_trustee_and_trust(self.osc, mock_cluster)

        mock_keystone.create_trustee.assert_called_once_with(
            '%s_%s' % (mock_cluster.uuid, mock_cluster.project_id),
            mock_password,
        )
        mock_keystone.create_trust.assert_called_once_with(
            mock_trustee.id,
        )
        self.assertEqual(mock_trustee.name, mock_cluster.trustee_username)
        self.assertEqual(mock_trustee.id, mock_cluster.trustee_user_id)
        self.assertEqual(mock_password, mock_cluster.trustee_password)
        self.assertEqual(mock_trust.id, mock_cluster.trust_id)

    @patch('magnum.common.utils.generate_password')
    def test_create_trustee_and_trust_with_error(self, mock_generate_password):
        mock_cluster = mock.MagicMock()
        mock_generate_password.side_effect = exception.MagnumException()

        self.assertRaises(exception.TrusteeOrTrustToClusterFailed,
                          trust_manager.create_trustee_and_trust,
                          self.osc,
                          mock_cluster)

    def test_delete_trustee_and_trust(self):
        mock_cluster = mock.MagicMock()
        mock_cluster.trust_id = 'trust_id'
        mock_cluster.trustee_user_id = 'trustee_user_id'
        mock_keystone = mock.MagicMock()
        self.osc.keystone.return_value = mock_keystone
        context = mock.MagicMock()

        trust_manager.delete_trustee_and_trust(self.osc, context,
                                               mock_cluster)

        mock_keystone.delete_trust.assert_called_once_with(
            context, mock_cluster
        )
        mock_keystone.delete_trustee.assert_called_once_with(
            'trustee_user_id',
        )

    def test_delete_trustee_and_trust_without_trust_id(self):
        mock_cluster = mock.MagicMock()
        mock_cluster.trust_id = None
        mock_cluster.trustee_user_id = 'trustee_user_id'
        mock_keystone = mock.MagicMock()
        self.osc.keystone.return_value = mock_keystone
        context = mock.MagicMock()

        trust_manager.delete_trustee_and_trust(self.osc, context,
                                               mock_cluster)

        self.assertEqual(0, mock_keystone.delete_trust.call_count)
        mock_keystone.delete_trustee.assert_called_once_with(
            'trustee_user_id',
        )

    def test_delete_trustee_and_trust_without_trustee_user_id(self):
        mock_cluster = mock.MagicMock()
        mock_cluster.trust_id = 'trust_id'
        mock_cluster.trustee_user_id = None
        mock_keystone = mock.MagicMock()
        self.osc.keystone.return_value = mock_keystone
        context = mock.MagicMock()

        trust_manager.delete_trustee_and_trust(self.osc, context, mock_cluster)

        mock_keystone.delete_trust.assert_called_once_with(
            context, mock_cluster
        )
        self.assertEqual(0, mock_keystone.delete_trustee.call_count)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webtest.app import AppError

from magnum.tests.unit.api import utils as apiutils
from magnum.tests.unit.common.policies import base
from magnum.tests.unit.objects import utils as obj_utils


class TestClusterPolicy(base.PolicyFunctionalTest):
    def setUp(self):
        super(TestClusterPolicy, self).setUp()
        self.cluster = obj_utils.create_test_cluster(
            self.context, name='cluster_example_A', node_count=3
        )

    def test_get_all_no_permission(self):
        exc = self.assertRaises(
            AppError, self.get_json, '/clusters',
            headers=self.member_headers)
        self.assertIn("403 Forbidden", str(exc))

    def test_get_no_permission(self):
        exc = self.assertRaises(
            AppError,
            self.get_json,
            f"/clusters/{self.cluster.uuid}",
            headers=self.member_headers)
        self.assertIn("403 Forbidden", str(exc))

    def test_create_no_permission(self):
        exc = self.assertRaises(
            AppError, self.post_json,
            '/clusters', apiutils.cluster_post_data(),
            headers=self.reader_headers)
        self.assertIn("403 Forbidden", str(exc))

    def test_update_no_permission(self):
        cluster_dict = [
            {'path': '/node_count', 'value': 4, 'op': 'replace'}
        ]
        exc = self.assertRaises(
            AppError, self.patch_json,
            f"/clusters/{self.cluster.name}", cluster_dict,
            headers=self.reader_headers
        )
        self.assertIn("403 Forbidden", str(exc))

    def test_delete_no_permission(self):
        # delete cluster
        exc = self.assertRaises(
            AppError, self.delete, f"/clusters/{self.cluster.uuid}",
            headers=self.reader_headers
        )
        self.assertIn("403 Forbidden", str(exc))
